package org.kaiwitte.workbench.data;

import net.tmorris.adt.sequence.Sequence;

import java.util.Date;

/**
 * A Factory to create {@link Condition Conditions}.
 */ 
public final class ConditionFactory {
    private ConditionFactory() {
        throw new UnsupportedOperationException();
    }

    private static final class ConditionImpl implements Condition {
        private final String str;
        public ConditionImpl(String str) {
            this.str = str;
        }
        public String toSQLString() {
            return str;
        }
        public String toString() {
            return str;
        }
    };

    /**
     * Returns a <code>Condition</code> that combines two conditions
     * with a logical AND.
     * @param c1 the first argument
     * @param c2 the second argument
     * @return the <code>Condition</code> c1 AND c2
     */ 
    public static Condition and(Condition c1, Condition c2) {
        return new ConditionImpl(c1.toSQLString() + " AND " + c2.toSQLString());
    }

    /**
     * Returns a <code>Condition</code> that requires a certain 
     * {@link DataColumn}'s {@link DataColumn#getName} to equal a certain
     * <code>String</code>.
     * @param column the {@link DataColumn} to be compared
     * @param s the <code>String</code> that is the comparator
     * @return the resulting <code>Condition</code>
     */ 
    public static Condition columnEqual(DataColumn column, String s) {
        if (!((column.getType().equals(String.class))
            || (column.getType().equals(String[].class)))) {
            throw new IllegalArgumentException("column is not of type String");
        }
        
        StringBuilder ret = new StringBuilder();
        ret.append(column.getName());
        ret.append("='");
        ret.append(s);
        ret.append("'");
        return new ConditionImpl(ret.toString());
    }

    /**
     * Returns a <code>Condition</code> that requires a certain 
     * {@link DataColumn}'s {@link DataColumn#getName} to equal a certain
     * <code>Date</code>.
     * @param column the {@link DataColumn} to be compared
     * @param d the <code>Date</code> that is the comparator
     * @return the resulting <code>Condition</code>
     */ 
    public static Condition columnEqual(DataColumn column, Date d) {
        if (!(column.getType().equals(Date.class))) {
            throw new IllegalArgumentException("column is not of type Date");
        }
        
        StringBuilder ret = new StringBuilder();
        ret.append(column.getName());
        ret.append("='");
        ret.append(new java.sql.Date(d.getTime()));
        ret.append("'");
        return new ConditionImpl(ret.toString());
    }

    /**
     * Returns a <code>Condition</code> that requires a certain 
     * {@link DataColumn}'s {@link DataColumn#getName} to equal a certain
     * <code>Boolean</code>.
     * @param column the {@link DataColumn} to be compared
     * @param b the <code>Boolean</code> that is the comparator
     * @return the resulting <code>Condition</code>
     */ 
    public static Condition columnEqual(DataColumn column, Boolean b) {
        if (!(column.getType().equals(Boolean.class))) {
            throw new IllegalArgumentException("column is not of type Boolean");
        }
        
        StringBuilder ret = new StringBuilder();
        ret.append(column.getName());
        ret.append("=");
        ret.append(b ? 1 : 0);
        return new ConditionImpl(ret.toString());
    }

    /**
     * Creates a {@link Condition} that allows any value for a certain 
     * {@link DataColumn}.
     * @param column the {@link DataColumn} to be compared 
     * @return the resulting {@link Condition}
     */ 
    public static Condition any(DataColumn column) {
        StringBuilder ret = new StringBuilder();
        ret.append(column.getName());
        ret.append(" regexp '");
        ret.append(".*");
        ret.append("'");
        return new ConditionImpl(ret.toString());
    }

    /**
     * Returns a <code>Condition</code> that requires a certain 
     * {@link DataColumn}'s {@link DataColumn#getName} to equal a certain
     * <code>long</code>.
     * @param column the {@link DataColumn} to be compared
     * @param i the <code>int</code> that is the comparator
     * @return the resulting <code>Condition</code>
     */ 
    public static Condition columnEqual(DataColumn column, long i) {
        if (!(column.getType().equals(Integer.class))) {
            throw new IllegalArgumentException("column is not of type Integer");
        }
        
        StringBuilder ret = new StringBuilder();
        ret.append(column.getName());
        ret.append("=");
        ret.append(i);
        return new ConditionImpl(ret.toString());
    }
    
    /**
     * Returns a <code>Condition</code> that requires a certain 
     * all text fields to start with the text in the text fields
     * of the provided {@link Record}.
     * @param r the {@link Record} to be compared
     * @return the resulting <code>Condition</code>
     */ 
    public static Condition textFieldsStartWith(Record r) {
        StringBuilder ret = new StringBuilder();
        
        DataScheme scheme = r.getScheme();
        Sequence<DataColumn> columns = scheme.getColumns();
        
        for (int i = 0; i < columns.getSize(); i++) {
            if (columns.get(i).getType().equals(String.class)
                    && ((String) r.getValue(i)).matches("\\w*")) {
                ret.append(columns.get(i).getName());
                ret.append(" REGEXP '^");
                ret.append((String) r.getValue(i));
                ret.append(".*'");
                ret.append(" AND ");
            }
        }
        
        String result = ret.toString();
        if (result.matches(".* AND ")) {
            result = result.substring(0, result.length() - 5);
        }
        return result == null ? null : new ConditionImpl(result);
    }
    
    /**
     * Returns a <code>Condition</code> from a <code>String</code>
     * that is in SQL syntax. E. g.
     * <pre>
     * name='kai' AND age<20
     * </pre>
     * @param s the <code>String</code> with the condition in SQL syntax
     * @return the resulting <code>Condition</code>
     */ 
    public static Condition rawSQLCondition(String s) {
        return new ConditionImpl(s);
    }
}
