package org.kaiwitte.workbench.data;

import net.tmorris.adt.sequence.Sequence;
import net.tmorris.adt.sequence.ArraySequenceImpl;

import java.io.IOException;
import static java.lang.System.*;
import java.sql.*;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.logging.Logger;

class DataManagerImpl implements DataManager {
    private static final Logger logger =
        Logger.getLogger("org.kaiwitte.workbench.sqlcommands");
    private static int instances;
    
    private String database;
    private DataScheme scheme;
    Statement stmt;
    
    private final Map<Class,String> types = new HashMap<Class,String>();
    {
        types.put(String.class, " CHAR(100)");
        types.put(String[].class, " CHAR(100)");
        types.put(Integer.class, " INT");
        types.put(Boolean.class, " BOOLEAN");
        types.put(java.util.Date.class, " DATE");
    }
    
    DataManagerImpl(String driver, String database, DataScheme scheme) 
            throws SQLException,
            ClassNotFoundException, IllegalAccessException, 
            InstantiationException {
        this.database = database;
        this.scheme = scheme;
        //out.println("DataManagerImpl instances: " + ++instances);

        Class.forName(driver).newInstance();
        Connection con = DriverManager.getConnection(database);
        stmt = con.createStatement();

        createTable(scheme, stmt);
    }
    
    private void createTable(DataScheme scheme, Statement stmt)
            throws SQLException {
        StringBuilder command = new StringBuilder();
        
        command.append("CREATE TABLE IF NOT EXISTS ");
        command.append(scheme.getName());
        command.append(" (id MEDIUMINT NOT NULL AUTO_INCREMENT, ");
        
        for (DataColumn column : scheme.getColumns()) {
            command.append(column.getName());
            command.append(types.get(column.getType()));
            command.append(", ");
        }
        
        command.append("PRIMARY KEY (id))");
        
        logger.info(command.toString());
        
        stmt.executeUpdate(command.toString());
    }

    public Sequence<Record> find(Condition condition) throws IOException {
        List<Record> results = new ArrayList<Record>();

        try {
            String query = createSearchQuery(condition);
            logger.info(query);
            ResultSet rs = stmt.executeQuery(query);
            while (rs.next()) {
                Record record = retrieveRecord(rs);
                results.add(record);
            }
        } catch (SQLException e) {
            throw new IOException(e.getMessage());
        }
        
        return ArraySequenceImpl.getArraySequence()
                .getSequence(results.toArray(new Record[0])); 
    }

    private String createSearchQuery(Condition condition) {
        StringBuilder ret = new StringBuilder();
        ret.append("SELECT * FROM ");
        ret.append(scheme.getName());
        ret.append(" WHERE ");
        ret.append(condition.toSQLString());
        return ret.toString();
    }

    private String createInsertQuery(Record record) {
        StringBuilder ret = new StringBuilder();
        ret.append("INSERT INTO ");
        ret.append(scheme.getName());

        ret.append("(");
        for (int i = 0; i < scheme.getColumns().getSize(); i++) {
            DataColumn column = scheme.getColumns().get(i);
            ret.append(column.getName());
            if (i < scheme.getLength() - 1) {
                ret.append(",");
            }
        }
        ret.append(")");

        ret.append(" VALUES (");
        for (int i = 0; i < scheme.getLength(); i++) {
            if (scheme.getColumns().get(i).getType().equals(String.class)
                || scheme.getColumns().get(i).getType().equals(java.util.Date.class)
                || scheme.getColumns().get(i).getType().equals(String[].class)) {
                ret.append("'");
            }
            DataColumn column = scheme.getColumns().get(i);
            if (column.getType().equals(java.util.Date.class)) {
                ret.append(
                    new java.sql.Date(
                        ((java.util.Date) record.getValue(i)).getTime()));
            }
            else {
                ret.append(record.getValue(i));
            }
            if (scheme.getColumns().get(i).getType().equals(String.class)
                || scheme.getColumns().get(i).getType().equals(java.util.Date.class)
                || scheme.getColumns().get(i).getType().equals(String[].class)) {
                ret.append("'");
            }
            if (i < scheme.getLength() - 1) {
                ret.append(",");
            }
        }
        ret.append(")");
        return ret.toString();
    }

    private Record retrieveRecord(ResultSet rs) throws SQLException {
        List fields = new ArrayList();
        for (DataColumn column : scheme.getColumns()) {
            FieldReader fieldReader = FIELD_READERS.get(column.getType());
            fields.add(fieldReader.readField(rs, column));
        }
        long id = rs.getInt("id");
        Record record = new MutableRecordImpl(id, scheme, fields.toArray());
        return record;
    }

    public void update(Record record) throws IOException {
        String command = createUpdateCommand(record);
        logger.info(command);

        try {
            stmt.executeUpdate(command);
        } catch (SQLException e) {
            throw new IOException(e.getMessage());
        }
    }

    private String createUpdateCommand(Record record) {
        StringBuilder command = new StringBuilder();
        command.append("UPDATE ");
        command.append(scheme.getName());
        command.append(" SET ");

        for (int i = 0; i < scheme.getLength(); i++) {
            DataColumn column = scheme.getColumns().get(i);
            command.append(column.getName());
            command.append("=");
            if (column.getType().equals(String.class)
                    || column.getType().equals(String[].class)
                    || column.getType().equals(java.util.Date.class)) {
                command.append("'");
            }
            if (column.getType().equals(java.util.Date.class)) {
                command.append(
                    new java.sql.Date(
                        ((java.util.Date) record.getValue(i)).getTime()));
            }
            else {
                command.append(record.getValue(i));
            }
            if (column.getType().equals(String.class)
                    || column.getType().equals(String[].class)
                    || column.getType().equals(java.util.Date.class)) {
                command.append("'");
            }
            if (i < scheme.getLength() - 1) {
                command.append(",");
            }
        }
        command.append(" WHERE ID=" + record.getID());
        
        return command.toString();
    }

    public void insert(Record record) throws IOException {
        String query = createInsertQuery(record);
        logger.info(query);
        try {
            stmt.executeUpdate(query);
        } catch (SQLException e) {
            throw new IOException(e.getMessage());
        }
    }
    
    public void delete(long recordNumber) throws IOException {
        String statement = createDeleteStatement(recordNumber);
        try {
            stmt.executeUpdate(statement);
        } catch (SQLException e) {
            throw new IOException(e.getMessage());
        }
    }

    private String createDeleteStatement(long recordNumber) {
        StringBuilder statement = new StringBuilder();
        statement.append("DELETE FROM ");
        statement.append(scheme.getName());
        statement.append(" WHERE ID=");
        statement.append(recordNumber);

        return statement.toString();
    }

    private static interface FieldReader {
        public Object readField(ResultSet rs, DataColumn column)
            throws SQLException;
    }
    private static final Map<Class,FieldReader> FIELD_READERS =
        new HashMap<Class,FieldReader>();
    static {
        FIELD_READERS.put(String.class, new FieldReader() {
            public Object readField(ResultSet rs, DataColumn column)
                   throws SQLException {
                return rs.getString(column.getName());
            }
        });
        FIELD_READERS.put(String[].class, new FieldReader() {
            public Object readField(ResultSet rs, DataColumn column)
                   throws SQLException {
                return rs.getString(column.getName());
            }
        });
        FIELD_READERS.put(java.util.Date.class, new FieldReader() {
            public Object readField(ResultSet rs, DataColumn column)
                   throws SQLException {
                return rs.getDate(column.getName());
            }
        });
        FIELD_READERS.put(Boolean.class, new FieldReader() {
            public Object readField(ResultSet rs, DataColumn column)
                   throws SQLException {
                return rs.getBoolean(column.getName());
            }
        });
        FIELD_READERS.put(Integer.class, new FieldReader() {
            public Object readField(ResultSet rs, DataColumn column)
                   throws SQLException {
                return rs.getInt(column.getName());
            }
        });
    }
}
