package org.kaiwitte.workbench.data;

import org.xml.sax.*;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.apache.xerces.parsers.DOMParser;

import java.io.IOException;
import java.io.InputStream;
import java.util.*;

import net.tmorris.adt.sequence.Sequence;
import net.tmorris.adt.sequence.ArraySequenceImpl;

public class DataSchemeFactoryImpl implements DataSchemeFactory {
    private MutableDataScheme scheme;
    private Map<String,MutableDataScheme> schemes =
        new HashMap<String,MutableDataScheme>();
    private List<String> schemesKeys = new ArrayList<String>();
    private static final String PRINT_TEMPLATE_SEPARATOR = ",";
    private final Map<String, MutableDataScheme> backupSchemeNameToSourceScheme
        = new HashMap<String, MutableDataScheme>();

    private static interface MutableDataColumn extends DataColumn {
        public void setIndirectlyHidden(boolean hidden);
        public void setTextAreaHeight(int i);
        public void setType(Class type);
        public void setRequiredForUpdate(boolean b);
        public void setName(String name);
        public void setAssociatedDataScheme(DataScheme scheme);
        public void setAssociatedDataSchemeName(String scheme);
        public void setNameInAssociatedDataScheme(String name);
        public String getAssociatedDataSchemeName();
        public void setChoices(String[] choices);
        public void setQualifiedName(String name);
        public void setHidden(boolean hidden);
        public void setSmall(boolean small);
    }

    private static interface MutableDataScheme extends DataScheme {
        public void addPrintTemplate(String[] template);
        public void addColumn(MutableDataColumn column);
        public Sequence<MutableDataColumn> getMutableDataColumns();
        public void setName(String name);
        public void setMailCondition(String name);

        public void setAutoLoadCondition(String s);
        public void setBackupScheme(DataScheme scheme);
    }

    private static class MutableDataSchemeImpl implements MutableDataScheme {
        private List<String[]> printTemplates =
            new ArrayList<String[]>();
        private List<MutableDataColumn> dataColumns =
            new ArrayList<MutableDataColumn>();
        private String name;
        private String mailCondition = "";
        private String autoLoadCondition = "";
        private DataScheme backupScheme;

        public int getLength() {
            return dataColumns.size();
        }

        public Sequence<DataColumn> getColumns() {
            return ArraySequenceImpl.getArraySequence()
                .getSequence(dataColumns.toArray(new DataColumn[0]));
        }

        public String getName() {
            return name;
        }

        public String getMailCondition() {
            return mailCondition;
        }

        public DataScheme getBackupScheme() {
            return backupScheme;
        }

        public String getAutoLoadCondition() {
            return autoLoadCondition;
        }

        public Iterator<String[]> printTemplates() {
            return Collections.unmodifiableList(printTemplates).iterator();
        }

        public String toString() {
            StringBuilder ret = new StringBuilder();
            for (DataColumn d : dataColumns) {
                ret.append(d);
                ret.append(" | ");
            }
            return ret.toString();
        }

        public void addPrintTemplate(String[] template) {
            printTemplates.add(template.clone());
        }

        public void addColumn(MutableDataColumn column) {
            dataColumns.add(column);
        }

        public Sequence<MutableDataColumn> getMutableDataColumns() {
            return ArraySequenceImpl.getArraySequence()
                .getSequence(dataColumns.toArray(new MutableDataColumn[0]));
        }

        public void setName(String name) {
            this.name = name;
        }

        public void setMailCondition(String mailCondition) {
            this.mailCondition = mailCondition;
        }

        public void setAutoLoadCondition(String s) {
            autoLoadCondition = s;
        }

        public void setBackupScheme(DataScheme scheme) {
            this.backupScheme = scheme;
        }
    };

    private static class MutableDataColumnImpl implements MutableDataColumn {
        private Class type;
        private boolean isRequiredForUpdate;
        private String name;
        private DataScheme scheme;
        private String nameInScheme;
        private String qualifiedName;

        public String associatedschemename;
        private String[] choices;
        private boolean hidden;
        private boolean small;
        private boolean indirectlyHidden;

        public void setIndirectlyHidden(boolean hidden) {
            this.indirectlyHidden = hidden;
        }

        public void setTextAreaHeight(int textAreaHeight) {
            this.textAreaHeight = textAreaHeight;
        }

        private int textAreaHeight;

        public int getTextAreaHeight() {
            return textAreaHeight;
        }

        public Class getType() {
            return type;
        }

        public boolean isRequiredForUpdate() {
            return isRequiredForUpdate;
        }

        public String getName() {
            return name;
        }

        public String getQualifiedName() {
            return qualifiedName;
        }

        public DataScheme getAssociatedDataScheme() {
            return scheme;
        }

        public String getNameInAssociatedDataScheme() {
            return nameInScheme;
        }

        public Sequence<String> getChoices() {
            return ArraySequenceImpl.getArraySequence()
                .getSequence(choices);
        }

        public boolean isHidden() {
            return hidden;
        }

        public boolean isSmall() {
            return small;
        }

        public boolean isIndirectlyHidden() {
            return indirectlyHidden;
        }

        public String toString() {
            return name + " " + type;
        }

        public void setType(Class type) {
            this.type = type;
        }

        public void setRequiredForUpdate(boolean b) {
            this.isRequiredForUpdate = b;
        }

        public void setName(String name) {
            this.name = name;
        }

        public void setAssociatedDataScheme(DataScheme scheme) {
            this.scheme = scheme;
        }

        public void setAssociatedDataSchemeName(String scheme) {
            this.associatedschemename = scheme;
        }

        public void setNameInAssociatedDataScheme(String name) {
            this.nameInScheme = name;
        }

        public String getAssociatedDataSchemeName() {
            return associatedschemename;
        }

        public void setChoices(String[] choices) {
            this.choices = choices.clone();
        }

        public void setQualifiedName(String qualifiedName) {
            this.qualifiedName = qualifiedName;
        }

        public void setHidden(boolean hidden) {
            this.hidden = true;
        }

        public void setSmall(boolean small) {
            this.small = small;
        }
    };

    private MutableDataColumn column;

    public Sequence<DataScheme> createDataScheme(InputStream in) throws IOException {
        if (in == null) {
            throw new NullPointerException();
        }

        DOMParser parser = new DOMParser();
        try {
            parser.parse(new InputSource(in));
            Document doc = parser.getDocument();
            Node topNode = doc.getDocumentElement();
            parseSchemes(topNode);
        } catch (SAXException e) {
            throw new IOException("error reading scheme: " + e);
        }
        List<DataScheme> ret = new ArrayList<DataScheme>();
        for (String key : schemesKeys) {

            // associate associated scheme
            MutableDataScheme returnScheme = schemes.get(key);
            for (MutableDataColumn column : returnScheme.getMutableDataColumns()) {
                column.setAssociatedDataScheme(
                    schemes.get(column.getAssociatedDataSchemeName()));
            }
            ret.add(returnScheme);
        }

        // add backup scheme
        for (String backupSchemeName:backupSchemeNameToSourceScheme.keySet()) {
            MutableDataScheme sourceScheme =
                backupSchemeNameToSourceScheme.get(backupSchemeName);
            DataScheme targetScheme = schemes.get(backupSchemeName);
            if (targetScheme == null) {
                throw new IOException(
                    "Scheme " + sourceScheme.getName() + " has specified "
                    + "associated Scheme " + backupSchemeName + ", but "
                    + "there is no such scheme specification in scheme.xml"
                );
            }
            sourceScheme.setBackupScheme(targetScheme);
        }

        return ArraySequenceImpl.getArraySequence()
                .getSequence(ret.toArray(new DataScheme[0]));
    }

    private void parseSchemes(Node node) throws IOException {
        NodeList children = node.getChildNodes();
        for (int i = 0; i < children.getLength(); i++) {
            if (children.item(i).getNodeName()
                     .equalsIgnoreCase("datascheme")) {
                scheme = new MutableDataSchemeImpl();
                parseScheme(children.item(i));
                schemes.put(scheme.getName(), scheme);
                schemesKeys.add(scheme.getName());
            }
        }
    }

    private void parseScheme(Node node) throws IOException {
        NodeList children = node.getChildNodes();
        for (int i = 0; i < children.getLength(); i++) {
            if (children.item(i).getNodeName().equalsIgnoreCase("name")) {
                scheme.setName(parseContent(children.item(i)));
            }
            else if (children.item(i).getNodeName()
                     .equalsIgnoreCase("datacolumn")) {
                parseDataColumn(children.item(i));
            }
            else if (children.item(i).getNodeName()
                     .equalsIgnoreCase("mailcondition")) {
                scheme.setMailCondition(parseContent(children.item(i)));
            }
            else if (children.item(i).getNodeName()
                     .equalsIgnoreCase("autoloadcondition")) {
                scheme.setAutoLoadCondition(parseContent(children.item(i)));
            }
            else if (children.item(i).getNodeName()
                     .equalsIgnoreCase("printtemplate")) {
                scheme.addPrintTemplate(
                    parseContent(children.item(i)).split(
                        PRINT_TEMPLATE_SEPARATOR));
            }
            else if (children.item(i).getNodeName()
                     .equalsIgnoreCase("backupdatascheme")) {
                String backupSchemeName = parseContent(children.item(i));
                backupSchemeNameToSourceScheme.put(backupSchemeName, scheme);
            }
        }
    }

    private void parseDataColumn(Node node) throws IOException {
        column = new MutableDataColumnImpl();
        NodeList children = node.getChildNodes();

        for (int i = 0; i < children.getLength(); i++) {
            if (children.item(i).getNodeName().equalsIgnoreCase("type")) {
                String content = parseContent(children.item(i));
                Class type = types.get(content);
                if (type == null) {
                    throw new IOException("invalid scheme format");
                }
                column.setType(type);
            }
            if (children.item(i).getNodeName().equalsIgnoreCase("columnname")) {
                String content = parseContent(children.item(i));
                column.setName(content);
            }
            if (children.item(i).getNodeName().equalsIgnoreCase("choices")) {
                String content = parseContent(children.item(i));
                column.setChoices(content.split(","));
            }
            if (children.item(i).getNodeName().equalsIgnoreCase("hidden")) {
                column.setHidden(true);
            }
            if (children.item(i).getNodeName().equalsIgnoreCase("small")) {
                column.setSmall(true);
            }
            if (children.item(i).getNodeName().equalsIgnoreCase(
                    "associateddatascheme")) {
                String content = parseContent(children.item(i));
                column.setAssociatedDataSchemeName(content);
            }
            if (children.item(i).getNodeName().equalsIgnoreCase(
                    "nameinassociatedscheme")) {
                String content = parseContent(children.item(i));
                column.setNameInAssociatedDataScheme(content);
            }
            if (children.item(i).getNodeName().equalsIgnoreCase("requiredforupdate")) {
                String content = parseContent(children.item(i));
                final boolean required = Boolean.parseBoolean(content);
                column.setRequiredForUpdate(required);
            }
            if (children.item(i).getNodeName().equalsIgnoreCase("indirectlyinvisible")) {
                String content = parseContent(children.item(i));
                final boolean required = Boolean.parseBoolean(content);
                column.setIndirectlyHidden(required);
            }
            if (children.item(i).getNodeName().equalsIgnoreCase("textareaheight")) {
                String content = parseContent(children.item(i));
                try {
                    column.setTextAreaHeight(
                        Integer.parseInt(content)
                    );
                } catch (NumberFormatException nfe) {
                    throw new IOException("invalid value for "
                                          + "textareaheight: "
                                          + content);
                }
            }
        }
        scheme.addColumn(column);
    }

    private String parseContent(Node node) {
        return node.getChildNodes().item(0).getNodeValue().trim();
    }

    private Map<String,Class> types = new HashMap<String,Class>();
    {
        types.put("int", Integer.class);
        types.put("string", String.class);
        types.put("string[]", String[].class);
        types.put("date", java.util.Date.class);
        types.put("boolean", Boolean.class);
    }
}
