package org.kaiwitte.workbench.data;

import net.tmorris.adt.sequence.Sequence;
import net.tmorris.adt.sequence.ArraySequenceImpl;

import java.io.IOException;
import java.util.logging.Logger;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.sql.*;
import static java.lang.System.*;

class MetaDataManagerImpl implements MetaDataManager {
    private static final Logger logger =
        Logger.getLogger("org.kaiwitte.workbench.sqlcommands");
    private DataManagerFactory dataManagerFactory;
    private static int instances;
    
    private String database;
    private MetaScheme metaScheme;
    Statement stmt;
    
    private String driver;

    private final Map<Class,String> types = new HashMap<Class,String>();
    private Connection connection = null;

    {
        types.put(String.class, " CHAR(100)");
        types.put(String[].class, " CHAR(100)");
        types.put(java.util.Date.class, " DATE");
        types.put(Integer.class, " INT");
        types.put(Boolean.class, " BOOLEAN");
    }
    
    MetaDataManagerImpl(String driver, String database, MetaScheme scheme) 
            throws IOException,
            ClassNotFoundException, IllegalAccessException, 
            InstantiationException {
        this.driver = driver;
        this.database = database;
        this.metaScheme = scheme;
        //out.println("MetaDataManagerImpl instances: " + ++instances);

        Class.forName(driver).newInstance();
        connect();
    }

    public void connect()
            throws IOException {
        try {
            connection = DriverManager.getConnection(database);
            stmt = connection.createStatement();

            dataManagerFactory = new DataManagerFactoryImpl(driver, database);

            createTables(metaScheme, stmt);
        } catch (SQLException e) {
            throw new IOException(e.toString());
        }
    }

    public boolean isConnected() throws IOException {//.
        try {
            return !connection.isClosed();
        } catch (SQLException e) {
            throw new IOException(e.toString());
        }
    }

    private void createTables(MetaScheme schemes, Statement stmt)
            throws SQLException {
        for (DataScheme scheme : schemes.getDataSchemes()) {
            StringBuilder command = new StringBuilder();

            command.append("CREATE TABLE IF NOT EXISTS ");
            command.append(scheme.getName());
            command.append(" (id MEDIUMINT NOT NULL AUTO_INCREMENT, ");

            for (DataColumn column : scheme.getColumns()) {
                command.append(column.getName());
                command.append(types.get(column.getType()));
                command.append(", ");
            }

            command.append("PRIMARY KEY (id))");

            logger.info(command.toString());

            stmt.executeUpdate(command.toString());
        }
    }

    public Sequence<MetaRecord> find(Condition condition) throws IOException {
        List<MetaRecord> results = new ArrayList<MetaRecord>();

        try {
            String query = createSearchQuery(condition);
            logger.info(query);
            ResultSet rs = stmt.executeQuery(query);
            while (rs.next()) {
                MetaRecord record = retrieveRecord(rs);
                results.add(record);
            }
        } catch (SQLException e) {
            throw new IOException(e.getMessage());
        }
        
        return ArraySequenceImpl.getArraySequence()
                .getSequence(results.toArray(new MetaRecord[0])); 
    }

    private String createSearchQuery(Condition condition) {
        StringBuilder ret = new StringBuilder();
        ret.append("SELECT * FROM ");
        Sequence<DataScheme> dataSchemes = this.metaScheme.getDataSchemes();
        for (int i = 0; i < dataSchemes.getSize(); i++) {
            ret.append(dataSchemes.get(i).getName());
            if (i < dataSchemes.getSize() - 1) {
                ret.append(", ");
            }
        }
        ret.append(" WHERE ");
        String queryString = extend(condition.toSQLString());
        ret.append(queryString);
        return ret.toString();
    }

    private String createSingleTableSearchQuery(Record r, Condition condition) {
        StringBuilder ret = new StringBuilder();
        ret.append("SELECT * FROM ");
        ret.append(r.getScheme().getName());
        
        ret.append(" WHERE ");
        ret.append(condition.toSQLString());
        return ret.toString();
    }

    private long acquireID(Record r) throws IOException {
        StringBuilder ret = new StringBuilder();
        ret.append("SELECT id FROM ");
        ret.append(r.getScheme().getName());
        ret.append(" WHERE ");
        ret.append(createCondition(r));
        String query = ret.toString();
        logger.info(query);
        try {
            ResultSet result;
            result = stmt.executeQuery(query);
            final boolean recordExists = result.next();
            if (!recordExists) {
                throw new IOException(
                    "unexpectedly missing record while retrieving id for record");
            }
            if (recordExists) {
                logger.finest("record found: " + r);
            }
            return result.getInt("id");
        } catch (SQLException e) {
            throw new IOException(e.getMessage());
        }
    }

    /**
     * Make sure that the query returns the right associated row from the
     * corresponding associated data schemes.
     * @param s a query string
     * @return an extended query string
     */ 
    private String extend(String s) {
        StringBuilder ret = new StringBuilder(s);
        for (DataColumn column : metaScheme.getDataSchemes().get(0).getColumns()) {
            String associatedName = column.getNameInAssociatedDataScheme();
            if (associatedName != null) {
                String associatedTable = 
                    column.getAssociatedDataScheme().getName();
                ret.append(" AND ");
                ret.append(associatedTable);
                ret.append(".");
                ret.append(associatedName);
                ret.append("=");
                ret.append(column.getName());
            }
        }
        return ret.toString();
    }

    private String createInsertQuery(Record record) {
        StringBuilder ret = new StringBuilder();
        ret.append("INSERT INTO ");
        ret.append(record.getScheme().getName());

        ret.append("(");
        for (int i = 0; i < record.getScheme().getColumns().getSize(); i++) {
            DataColumn column = record.getScheme().getColumns().get(i);
            ret.append(column.getName());
            if (i < record.getScheme().getLength() - 1) {
                ret.append(",");
            }
        }
        ret.append(")");

        ret.append(" VALUES (");
        for (int i = 0; i < record.getScheme().getLength(); i++) {
            ret.append(QUOTES.get(record.getScheme().getColumns().get(i).getType()));
            DataColumn column = record.getScheme().getColumns().get(i);
            if (record.getValue(i).getClass().equals(java.util.Date.class)) {
                ret.append(new java.sql.Date(
                    ((java.util.Date) record.getValue(i)).getTime()));
            }
            else if (record.getValue(i).getClass().equals(Boolean.class)) {
                ret.append(
                    ((Boolean) record.getValue(i)) ? 1 : 0);
            }
            else if (column.getType().equals(java.util.Date.class)) {
                ret.append(
                    new java.sql.Date(
                        ((java.util.Date) record.getValue(i)).getTime()));
            }
            else {
                ret.append(record.getValue(i));
            }
            ret.append(QUOTES.get(record.getScheme().getColumns().get(i).getType()));
            if (i < record.getScheme().getLength() - 1) {
                ret.append(",");
            }
        }
        ret.append(")");
        return ret.toString();
    }

    private MetaRecord retrieveRecord(ResultSet rs) throws SQLException {
        List<Record> records = new ArrayList<Record>();
        for (DataScheme scheme : metaScheme.getDataSchemes()) {
            List fields = new ArrayList();
            for (DataColumn column : scheme.getColumns()) {
                FieldReader fieldReader = FIELD_READERS.get(column.getType());
                fields.add(fieldReader.readField(rs, scheme.getName(), column));
            }
            long id = rs.getInt(scheme.getName() + ".id");
            Record record = new MutableRecordImpl(id, scheme, fields.toArray());
            records.add(record);
        }
        return new MetaRecordImpl(records.toArray(new Record[0]));
    }

    public void update(MetaRecord mr) throws IOException {
        for (Record record : mr.getRecords()) {
            String command = createUpdateCommand(record);

            try {
                logger.info(command);
                stmt.executeUpdate(command);
            } catch (SQLException e) {
                throw new IOException(e.getMessage());
            }
        }
    }

    private String createUpdateCommand(Record record) {
        StringBuilder command = new StringBuilder();
        command.append("UPDATE ");
        command.append(record.getScheme().getName());
        command.append(" SET ");

        for (int i = 0; i < record.getScheme().getLength(); i++) {
            DataColumn column = record.getScheme().getColumns().get(i);
            command.append(column.getName());
            command.append("=");
            command.append(
                QUOTES.get(record.getScheme().getColumns().get(i).getType()));
            if (column.getType().equals(java.util.Date.class)) {
                command.append(
                    new java.sql.Date(
                        ((java.util.Date) record.getValue(i)).getTime()));
            }
            else {
                command.append(record.getValue(i));
            }
            command.append(
                QUOTES.get(record.getScheme().getColumns().get(i).getType()));
            if (i < record.getScheme().getLength() - 1) {
                command.append(",");
            }
        }
        command.append(" WHERE ID=" + record.getID());
        
        return command.toString();
    }

    public MetaRecord insert(MetaRecord meta) throws IOException {
        Record primaryRecord = meta.getRecords().get(0);
        List<Record> records = new ArrayList<Record>();
        List<Long> insertIDs = new ArrayList<Long>();
        
        /*
        * handle secondary records
        */
        for (int i = 1; i < meta.getRecords().getSize(); i++) {
            Record record = meta.getRecords().get(i);
            String query = createInsertQuery(record);
            try {
                if (!recordExists(record)) {
                    logger.info(query);
                    stmt.executeUpdate(query);
                    insertIDs.add((long) acquireLastID());
                }
                else {
                    insertIDs.add(acquireID(record));
                }
                Condition c = createCondition(record);
                DataManager dm = 
                    dataManagerFactory.getDataManager(record.getScheme());
//                    new DataManagerImpl(
//                    driver, database, record.getScheme());
                Record idRecord = dm.find(c).get(0);
                records.add(idRecord);
            } catch (SQLException e) {
                throw new IOException(e.getMessage());
            }
        }
        
        int j = 0;
        Sequence<DataColumn> columns = primaryRecord.getScheme().getColumns();
        List fields = new ArrayList();
        for (int i = 0; i < columns.getSize(); i++) {
            DataColumn column = columns.get(i);
            if (column.getAssociatedDataScheme() != null) {
                fields.add(records.get(j++).getID());
            }
            else {
                fields.add(primaryRecord.getValue(i));
            }
        }
        
        Record newPrimaryRecord = new MutableRecordImpl(
            primaryRecord.getScheme(), fields.toArray());
        
        String query = createInsertQuery(newPrimaryRecord);
        logger.info(query);
        try {
            stmt.executeUpdate(query);
        } catch (SQLException e) {
            throw new IOException(e.getMessage());
        }
        
        insertIDs.add(0, acquireLastID());
        
        MutableMetaRecord ret = new MutableMetaRecordImpl(meta);
        //out.println(insertIDs);
        ret.setID(insertIDs.toArray(new Long[0]));

        return ret;
    }

    private long acquireLastID() throws IOException {
        int id = 0;
        String query = "SELECT LAST_INSERT_ID()";
        logger.info(query);
        try {
            ResultSet rs = stmt.executeQuery(query);
            if (!rs.next()) {
                throw new IOException(
                    "the database did not assign an ID to the last inserted "
                    + "record for unknown reasons"
                );
            }
            id = rs.getInt("last_insert_id()");
            if (rs.next()) {
                throw new IOException(
                    "the database has assigned too many IDs to the last "
                    + "inserted "
                    + "record for unknown reasons"
                );
            }
        } catch (SQLException e) {
            throw new IOException(e.getMessage());
        }
        return id;
    }

    public void oldinsert(MetaRecord meta) throws IOException {
        for (Record record : meta.getRecords()) {
            String query = createInsertQuery(record);
            logger.info(query);
            try {
                if (!recordExists(record)) {
                    stmt.executeUpdate(query);
                }
            } catch (SQLException e) {
                throw new IOException(e.getMessage());
            }
        }
    }

    private boolean recordExists(Record record) throws IOException {
        Condition c = createCondition(record);
        String query = createSingleTableSearchQuery(record, c);
        logger.info(query);
        try {
            ResultSet result;
            result = stmt.executeQuery(query);
            final boolean recordExists = result.next();
            if (recordExists) {
                logger.finest("record exits already: " + record);
            }
            else {
                logger.finest("no such record: " + record);
            }
            return recordExists;
        } catch (SQLException e) {
            throw new IOException(e.getMessage());
        }
    }

    private Condition createCondition(Record record) {
        Condition ret = null;
        for (int i = 0; i < record.getScheme().getLength(); i++) {
            DataColumn dataColumn = record.getScheme().getColumns().get(i);
            Class type = dataColumn.getType();
            if (type.equals(Integer.class)) {
                Condition c = ConditionFactory.columnEqual(
                    dataColumn, (Integer) record.getValue(i)
                );
                if (ret == null) {
                    ret = c;
                }
                else {
                    ret = ConditionFactory.and(ret, c);
                }
            }
            if (type.equals(String.class)) {
                Condition c = ConditionFactory.columnEqual(
                    dataColumn, (String) record.getValue(i)
                );
                if (ret == null) {
                    ret = c;
                }
                else {
                    ret = ConditionFactory.and(ret, c);
                }
            }
            if (type.equals(String[].class)) {
                Condition c = ConditionFactory.columnEqual(
                    dataColumn, (String) record.getValue(i)
                );
                if (ret == null) {
                    ret = c;
                }
                else {
                    ret = ConditionFactory.and(ret, c);
                }
            }
            if (type.equals(java.util.Date.class)) {
                Condition c = ConditionFactory.columnEqual(
                    dataColumn, (java.util.Date) record.getValue(i)
                );
                if (ret == null) {
                    ret = c;
                }
                else {
                    ret = ConditionFactory.and(ret, c);
                }
            }
            if (type.equals(Boolean.class)) {
                Condition c = ConditionFactory.columnEqual(
                    dataColumn, (Boolean) record.getValue(i)
                );
                if (ret == null) {
                    ret = c;
                }
                else {
                    ret = ConditionFactory.and(ret, c);
                }
            }
        }
        return ret;
    }

    public void delete(long recordNumber) throws IOException {
        String statement = createDeleteStatement(recordNumber);
        logger.info(statement);
        try {
            stmt.executeUpdate(statement);
        } catch (SQLException e) {
            throw new IOException(e.getMessage());
        }
    }

    private String createDeleteStatement(long recordNumber) {
        StringBuilder statement = new StringBuilder();
        statement.append("DELETE FROM ");
        statement.append(metaScheme.getDataSchemes().get(0).getName());
        statement.append(" WHERE ID=");
        statement.append(recordNumber);

        return statement.toString();
    }

    private static interface FieldReader {
        public Object readField(ResultSet rs, String table, DataColumn column)
            throws SQLException;
    }
    private static final Map<Class,FieldReader> FIELD_READERS =
        new HashMap<Class,FieldReader>();
    static {
        FIELD_READERS.put(String.class, new FieldReader() {
            public Object readField(ResultSet rs, String table, 
                                    DataColumn column)
                   throws SQLException {
                return rs.getString(table + "." + column.getName());
            }
        });
        FIELD_READERS.put(String[].class, new FieldReader() {
            public Object readField(ResultSet rs, String table, 
                                    DataColumn column)
                   throws SQLException {
                return rs.getString(table + "." + column.getName());
            }
        });
        FIELD_READERS.put(java.util.Date.class, new FieldReader() {
            public Object readField(ResultSet rs, String table, 
                                    DataColumn column)
                   throws SQLException {
                return rs.getDate(table + "." + column.getName());
            }
        });
        FIELD_READERS.put(Boolean.class, new FieldReader() {
            public Object readField(ResultSet rs, String table, 
                                    DataColumn column)
                   throws SQLException {
                return rs.getBoolean(table + "." + column.getName());
            }
        });
        FIELD_READERS.put(Integer.class, new FieldReader() {
            public Object readField(ResultSet rs, String table, 
                                    DataColumn column)
                   throws SQLException {
                return rs.getInt(table + "." + column.getName());
            }
        });
    }
    private static final Map<Class,String> QUOTES =
        new HashMap<Class,String>();
    static {
        QUOTES.put(String.class, "'");
        QUOTES.put(String[].class, "'");
        QUOTES.put(java.util.Date.class, "'");
        QUOTES.put(Integer.class, "");
        QUOTES.put(Boolean.class, "");
    }
}
