package org.kaiwitte.workbench.tools;

import static java.lang.System.*;
import org.kaiwitte.workbench.data.*;
import org.kaiwitte.workbench.ui.UIConfig;
import org.kaiwitte.workbench.ui.controller.IconCreator;
import org.kaiwitte.workbench.ui.controller.ColorCellRenderer;

import javax.swing.*;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.util.Map;
import java.util.HashMap;
import java.io.*;
import java.net.MalformedURLException;
import java.awt.*;

import net.tmorris.adt.sequence.Sequence;

/**
 * A class to provide general tools for the project.
 */
public final class Tools {
    private static String NEWLINE =
        System.getProperty("line.separator", "\n");
    private static final int DATE_ID_LENGTH = 4;
    private static final int ID100_LENGTH = 5;
    private static final int ID100_OFFSET = 100;

    private Tools() {
    }

    /**
     * Replaces all variables with the format
     * <pre>
     * {recordname.columnname}
     * </pre>
     * in <code>content</code>with the according value in 
     * <code>metaRecord</code>.
     * @param metaRecord the record with the values
     * @param content the template text
     * @return the modified text
     */
    public static String replaceVariables(MetaRecord metaRecord,
                                          String content,
                                          UIConfig config) throws IOException {
        Sequence<Record> records = metaRecord.getRecords();
        for (Record record : records) {
            DataScheme scheme = record.getScheme();
            Sequence<DataColumn> columns = scheme.getColumns();
            for (int i = 0; i < columns.getSize(); i++) {
                StringBuilder variableName = new StringBuilder();

                if (isDateColumn(columns.get(i))) {
                    MetaDataManager data =
                        config.getMetaDataManager(
                            MetaSchemeFactory.getMetaScheme(
                                metaRecord.getRecords().get(0).getScheme())
                        );
                    content = content.replaceAll(Pattern.quote("{dateid}"),
                                                 computeDateID(data,
                                                               columns.get(i)));
                }

                // e. g.: {Kunden.Name}
                variableName.append("{");
                variableName.append(scheme.getName());
                variableName.append(".");
                variableName.append(columns.get(i).getName());
                variableName.append("}");

                // e. g.: {Kunden.Name} -> Kai Witte
                content = content.replaceAll(
                    Pattern.quote(variableName.toString()),
                    record.getValue(i).toString());
            }

            // insert ID
            //out.println(Pattern.quote(
            //        "{" + scheme.getName() + ".ID}"));

            String id100 = computeID100(record.getID());
            content = content.replaceAll(
                Pattern.quote("{" + scheme.getName() + ".ID100}"),
                id100
            );

            content = content.replaceAll(
                Pattern.quote(
                    "{" + scheme.getName() + ".ID}"),
                Long.toString(record.getID())
            );
        }

        // insert date
        content = content.replaceAll(
            Pattern.quote("{date}"),
            new java.util.Date().toString());

        // make images valid
//        Pattern p = Pattern.compile("<img\\s+src\\s*=\\s*\"(.*)\"\\s*>");
//        Matcher m = p.matcher(content);
//        if (m.find()) {
//            String filename = m.group(1);
//            out.println(filename);
//            String url;
//            try {
//                url = new File(filename).toURI().toURL().toString();
//                out.println(url);
//            } catch (MalformedURLException e) { config.getMainModel().logException(e);
//                throw new IOException(e.getMessage());
//            }
//            content = content.replaceAll(
//                filename,
//                Pattern.quote(url)
//            );
//        }


        return content;
    }

    private static String computeID100(long id) {
        id += ID100_OFFSET;
        String ret = Long.toString(id);

        // append leading 0s
        while (ret.length() < ID100_LENGTH) {
            ret = "0" + ret;
        }

        return ret;
    }

    private static String computeDateID(MetaDataManager data, DataColumn dc)
            throws IOException {
        Condition today = ConditionFactory.columnEqual(dc,
                                                       new java.util.Date());
        Sequence<MetaRecord> mr = data.find(today);
        String ret = Integer.toString(mr.getSize() - 1);

        while (ret.length() < DATE_ID_LENGTH) {
            ret = "0" + ret;
        }

        return ret;
    }

    private static boolean isDateColumn(DataColumn column) {
        return column.getName().toLowerCase().contains("date")
               || column.getName().toLowerCase().contains("datum");
    }

    /**
     * Reads a resouce to a <code>String</code>.
     * @param resource the source
     * @return the content
     * @throws IOException iff an I/O error occurs
     */
    public static String readResource(String resource) throws IOException {
        InputStream is = Tools.class.getResourceAsStream(resource);
        BufferedReader reader = new BufferedReader(
            new InputStreamReader(is));

        String line = null;
        StringBuilder ret = new StringBuilder();
        while ((line = reader.readLine()) != null) {
            ret.append(line);
            ret.append(NEWLINE);
        }

        return ret.toString();
    }

    /**
     * Reads a file to a <code>String</code>.
     * @param filename the name of the source file
     * @return the content
     * @throws IOException iff an I/O error occurs
     */
    public static String readFile(String filename) throws IOException {
        InputStream is = new FileInputStream(filename);
        StringBuilder ret;
        try {
            BufferedReader reader = new BufferedReader(
                new InputStreamReader(is));

            String line = null;
            ret = new StringBuilder();
            while ((line = reader.readLine()) != null) {
                ret.append(line);
                ret.append(NEWLINE);
            }
        } finally {
            is.close();
        }

        return ret.toString();
    }

    public static IconCreator iconCreator(final UIConfig config) {
        final Map<String,Icon> map = new HashMap<String,Icon>();

        map.put(config.getOption("red"), ColorCellRenderer.RED);
        map.put(config.getOption("blue"), ColorCellRenderer.BLUE);
        map.put(config.getOption("green"), ColorCellRenderer.GREEN);
        map.put(config.getOption("yellow"), ColorCellRenderer.YELLOW);
        map.put(config.getOption("white"), ColorCellRenderer.WHITE);
        map.put(config.getOption("orange"), ColorCellRenderer.ORANGE);
        map.put(config.getOption("pink"), ColorCellRenderer.PINK);
        map.put(config.getOption("purple"), ColorCellRenderer.PURPLE);
        map.put(config.getOption("cyan"), ColorCellRenderer.CYAN);
        map.put(config.getOption("lightblue"), ColorCellRenderer.LIGHTBLUE);

        return new IconCreator() {
            public Icon getIconFor(Object o) {
                Icon icon = map.get(o.toString());
                return icon;
            }
        };
    }
    
    public static <T> T[] sequenceToArray(Sequence<T> s, Class c) {
        if (c.getComponentType() == null) {
            throw new IllegalArgumentException(
                "Class parameter must be an array class type "
                + "like String[].class");
        }
        T[] ret = (T[]) java.lang.reflect.Array.
		newInstance(c.getComponentType(), s.getSize());
        int i = 0;
        for (T element : s) {
            ret[i++] = element;
        }
        return ret;
    }
}
