package org.kaiwitte.workbench.ui.model;

import org.kaiwitte.workbench.ui.view.MainView;
import org.kaiwitte.workbench.ui.view.SchemeTablePanel;
import org.kaiwitte.workbench.ui.UIConfig;
import org.kaiwitte.workbench.data.*;
import org.kaiwitte.joptions.Options;
import org.kaiwitte.joptions.OptionManagerFactory;
import org.kaiwitte.joptions.OptionManager;
import org.kaiwitte.tools.SwingTools;

import javax.swing.*;
import java.io.*;
import java.util.*;
import java.util.Timer;
import java.util.logging.Logger;
import java.util.logging.StreamHandler;
import java.util.logging.SimpleFormatter;
import java.util.logging.Level;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;

import net.tmorris.adt.sequence.Sequence;
import net.tmorris.adt.sequence.ArraySequenceImpl;

/**
 * The class to start the application and initialise other components.
 */
public final class MainModel implements Runnable {
    private static final boolean TRIAL_MODE = false;
    private static final long TRIAL_EXPIRATION = 1158326320421L;
    private UIConfig config;
    private final StringOutputStream out = new StringOutputStream();
    //private final StringOutputStream err = new StringOutputStream();
    private final StringOutputStream log = new StringOutputStream();

    private final Logger throwableLogger =
        Logger.getLogger("org.kwitte.workbench.exceptions");
    private final Logger generalLogger = Logger.global;
    private final boolean redirectOut;
    private static final int TRIAL_TIME = 120000;

    private MainModel(boolean redirectOut) {
        this.redirectOut = redirectOut;
        if (redirectOut) {
            //System.out.println("red");
            redirectOutput();
        }
        generalLogger.addHandler(new StreamHandler(out, new SimpleFormatter()));
        generalLogger.setLevel(Level.ALL);
        throwableLogger.addHandler(
            new StreamHandler(out, new SimpleFormatter()));
        throwableLogger.setLevel(Level.ALL);
    }

    private void redirectOutput() {
        //System.out.println("redirecting");
        System.setOut(new PrintStream(this.out));
        System.setErr(new PrintStream(this.out));
    }

    public void logException(Throwable t) {
        throwableLogger.info(errorText(t));
        //throwableLogger.throwing("MainModel", "logException", t);
    }

    public String getOutData() {
        return out.toString();
    }

    public String getLogData() {
        return log.toString();
    }

    public static void main(String[] args) {
        setUncaughtExceptionHandler();
        SwingUtilities.invokeLater(new MainModel(args.length == 0));
    }

    private static void setUncaughtExceptionHandler() {
        Thread.setDefaultUncaughtExceptionHandler(
            new Thread.UncaughtExceptionHandler() {
                public void uncaughtException(Thread t, Throwable e) {
                    e.printStackTrace();
                    SwingTools.exceptionDialog(
                                        "<html>"
                                        + "Unexpected internal program error."
                                        + "<br>"
                                        + "Please file a report.</html>",
                                        e);
                }
            }
        );
    }

    public void run() {
        UIConfigImpl config = new UIConfigImpl();
        MainView view = new MainView(config);
        this.config = config;
        MainModel mainModel = this;
        config.view = view;
        config.model = mainModel;

        Sequence<DataScheme> schemes;
        OptionManager manager = readInternalConfiguration(config, view);

        schemes = readExternalConfiguration(config);

        for (DataScheme scheme : schemes) {
            MetaScheme metaScheme = MetaSchemeFactory.getMetaScheme(scheme);

            MetaSchemeTableModel model =
                MetaSchemeTableModelFactory.getMetaSchemeTableModel(metaScheme);
            config.models.add(model);

            try {
                loadDefaultRecords(metaScheme, model, config);
            } catch (IOException e) {
                config.getMainModel().logException(e);
                final DataScheme dataScheme =
                    metaScheme.getDataSchemes().get(0);
                SwingTools.exceptionDialog(
                    "error loading default records "
                    + "for scheme " + dataScheme.getName()
                    + "; <autoloadcondition>: " +
                    dataScheme.getAutoLoadCondition(),
                    e);

                /*
                * Try it again with correct database settings:
                */
                int choice =
                    JOptionPane.showConfirmDialog(
                        null,
                        "Change database settings and try again?",
                        "database error",
                        JOptionPane.YES_NO_OPTION,
                        JOptionPane.QUESTION_MESSAGE
                    );
                if (choice == JOptionPane.YES_OPTION) {
                    manager.showDialog();
                    run();
                    return;
                }
            }

            SchemeTablePanel panel =
                new SchemeTablePanel(config, model);

            view.addTab(metaScheme.getName(), panel);
        }
        if (TRIAL_MODE) initialiseTrialMode(view);
        view.pack();
        view.setVisible(true);
    }

    private Sequence<DataScheme> readExternalConfiguration(
            UIConfigImpl config) {
        Sequence<DataScheme> schemes = null;
        try {
            /*
             * Dataschemes
             */
            InputStream in = new FileInputStream(
                                 "options/scheme.xml");
            if (in == null) {
                JOptionPane.showMessageDialog(
                    null,
                    "file missing: scheme.xml",
                    "error",
                    JOptionPane.ERROR_MESSAGE
                );
                System.exit(1);
            }
            schemes =
                new DataSchemeFactoryImpl().createDataScheme(in);
        } catch (IOException e) {
            config.getMainModel().logException(e);
            SwingTools.exceptionDialog(
                "error reading file: scheme.xml",
                e);
            System.exit(1);
        }

        MetaDataManagerFactory factory =
            new MetaDataManagerFactoryImpl(
                config.getOption("driver"),
                "jdbc:mysql://"
                + config.getOption("host") + "/"
                + config.getOption("databasename")+ "?user="
                + config.getOption("username") + "&password="
                + config.getOption("password")
            );
        config.metaFactory = factory;

        DataManagerFactory dmFactory =
            new DataManagerFactoryImpl(
                config.getOption("driver"),
                "jdbc:mysql://"
                + config.getOption("host") + "/"
                + config.getOption("databasename")+ "?user="
                + config.getOption("username") + "&password="
                + config.getOption("password")
            );
        config.factory = dmFactory;
        if (schemes == null) throw new AssertionError();
        return schemes;
    }

    private OptionManager readInternalConfiguration(UIConfigImpl config,
                                                    MainView view) {
        OptionManager manager;
        try {
            /*
             * initialize options
             */
            InputStream ino =
                MainModel.class.getResourceAsStream("/res/options.xml");
            if (ino == null) {
                JOptionPane.showMessageDialog(
                    null,
                    "file missing: options.xml",
                    "error",
                    JOptionPane.ERROR_MESSAGE
                );
                System.exit(1);
            }
            ResourceBundle optionBundle =
                ResourceBundle.getBundle("res/options");
            manager =
                OptionManagerFactory.createOptionManager(
                    ino, MainModel.class, optionBundle);
            config.options = manager.getOptions();
            view.setOptionManager(manager);
        } catch (IOException e) {
            config.getMainModel().logException(e);
            throw new AssertionError(e);
        }
        return manager;
    }

    private void loadDefaultRecords(MetaScheme metaScheme,
                                    MetaSchemeTableModel model,
                                    UIConfigImpl config)
            throws IOException {
        String conditionStr =
            metaScheme.getDataSchemes().get(0).getAutoLoadCondition();
        if (conditionStr.length() != 0) {
            MetaDataManager manager = config.getMetaDataManager(metaScheme);

            Condition c = ConditionFactory.rawSQLCondition(conditionStr);
            Sequence<MetaRecord> records = manager.find(c);

            for (MetaRecord record : records) {
                model.addRecord(record);
            }
        }
    }

    private String errorText(Throwable error) {
        final StringWriter output = new StringWriter();
        PrintWriter writer = new PrintWriter(output);

        error.printStackTrace(writer);

        return output.toString();
    }

    private static class UIConfigImpl implements UIConfig {
        private MainView view;
        private MainModel model;
        private MetaDataManagerFactory metaFactory;
        private Options options;
        private DataManagerFactory factory;
        private List<MetaSchemeTableModel> models =
            new ArrayList<MetaSchemeTableModel>();
        ResourceBundle resourceBundle =
            ResourceBundle.getBundle("res/resources");

        public MainView getMainView() {
            return view;
        }
        public MainModel getMainModel() {
            return model;
        }

        public MetaDataManager getMetaDataManager(MetaScheme scheme)
                throws IOException {
            return metaFactory.getMetaDataManager(scheme);
        }
        
        public MetaDataManagerFactory getMetaDataManagerFactory() {
            return metaFactory;
        }

        public Sequence<MetaSchemeTableModel> getMetaSchemeTableModels() {
            return
                ArraySequenceImpl.getArraySequence()
                .getSequence(
                models.toArray(new MetaSchemeTableModel[0]));
        }

        public String getResource(String key) {
            try {
                return resourceBundle.getString(key);
            } catch (MissingResourceException mre) {
                if (view == null) {
                    throw new AssertionError(mre);
                }
                else {
                    view.error("internal program error - please report to supplier",
                               mre);
                    return key;
                }
            }
        }

        public DataManager getDataManager(DataScheme scheme)
                throws IOException {
            return factory.getDataManager(scheme);
        }

        public String getOption(String key) {
            return options.getOption(key).toString();
        }
    }
    
    private void initialiseTrialMode(final MainView m) {
        if (System.currentTimeMillis() > TRIAL_EXPIRATION) {
            JOptionPane.showMessageDialog(
                m,
                "trial mode - expired",
                "trial",
                JOptionPane.INFORMATION_MESSAGE
            );
            System.exit(0);
        }
        ActionListener task = new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                JOptionPane.showMessageDialog(
                    m,
                    "trial mode - exiting after 2 minutes",
                    "trial",
                    JOptionPane.INFORMATION_MESSAGE
                );
                System.exit(0);
            }
        };
        javax.swing.Timer timer = 
            new javax.swing.Timer(TRIAL_TIME, task);
        timer.start();
    }
}
