package org.kaiwitte.workbench.ui.model;

import org.kaiwitte.workbench.data.MetaScheme;
import org.kaiwitte.workbench.data.MetaRecord;
import org.kaiwitte.workbench.data.DataColumn;

import javax.swing.table.AbstractTableModel;
import static java.lang.System.*;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;

import net.tmorris.adt.sequence.Sequence;

public class MetaSchemeTableModelFactory {
    private static final Map<MetaScheme,MetaSchemeTableModel> models =
        new HashMap<MetaScheme, MetaSchemeTableModel>();

    private MetaSchemeTableModelFactory() {
        throw new UnsupportedOperationException();
    }

    public static synchronized MetaSchemeTableModel getMetaSchemeTableModel(
            MetaScheme scheme) {
        MetaSchemeTableModel ret = models.get(scheme);
        if (ret == null) {
            ret = new MetaSchemeTableModelImpl(scheme);
            models.put(scheme, ret);
        }
        return ret;
    }

    /**
     * A <code>TableModel</code> to hold {@link org.kaiwitte.workbench.data.MetaRecord}s.
     */
    private static final class MetaSchemeTableModelImpl
            extends AbstractTableModel implements MetaSchemeTableModel {
        private List<MetaRecord> rows = new ArrayList<MetaRecord>();
        private final MetaScheme metaScheme;

        /**
         * Creates a new instance.
         * @param scheme the <code>MetaScheme</code> that defines the
         * structure of this <code>TableModel</code>
         * @throws NullPointerException iff <code>scheme == null</code>
         */
        MetaSchemeTableModelImpl(MetaScheme scheme) {
            if (scheme == null) {
                throw new NullPointerException();
            }
            this.metaScheme = scheme;
        }

        /**
         * Adds a record as a table row.
         * @param r the new record
         */
        public void addRecord(Sequence<MetaRecord> r) {
            for (MetaRecord record : r) {
                rows.add(record);
            }
            fireTableDataChanged();
        }

        /**
         * Adds a record as a table row.
         * @param r the new record
         */
        public void addRecord(MetaRecord... r) {
            for (MetaRecord record : r) {
                rows.add(record);
            }
            fireTableDataChanged();
        }

        @Override
        public String getColumnName(int column) {
            int virtualColumn = 0;
            for (int i = 0; i < metaScheme.getColumns().getSize(); i++) {
                while (metaScheme.getColumns().get(i).isHidden()) {
                    if (i >= metaScheme.getColumns().getSize()) {
                        throw new IndexOutOfBoundsException();
                    }
                    i++;
                }
                if (virtualColumn == column) {
                    return metaScheme.getColumns().get(i).getName();
                }
                virtualColumn++;
            }
            throw new IndexOutOfBoundsException();
        }

        @Override
        public Class<?> getColumnClass(int column) {
            int virtualColumn = 0;
            for (int i = 0; i < metaScheme.getColumns().getSize(); i++) {
                while (metaScheme.getColumns().get(i).isHidden()) {
                    if (i >= metaScheme.getColumns().getSize()) {
                        throw new IndexOutOfBoundsException();
                    }
                    i++;
                }
                if (virtualColumn == column) {
                    Class type = metaScheme.getColumns().get(i).getType();
                    return type.equals(String[].class) ? String.class : type;
                }
                virtualColumn++;
            }
            throw new IndexOutOfBoundsException();
        }

        /**
         * Removes all entries.
         */
        public void clear() {
            rows.clear();
            fireTableDataChanged();
        }

        /**
         * Returns the scheme that defines this table model's structure.
         * @return the scheme 
         */
        public MetaScheme getScheme() {
            return metaScheme;
        }

        /**
         * Returns the <code>MetaRecord</code> in the specified row.
         * @param row the row number, 0-based
         * @return the <code>MetaRecord</code> in the specified row
         * @throws IndexOutOfBoundsException iff <code>row < 0</code>
         * or <code>row ></code> the maximum index  
         */
        public MetaRecord getRecord(int row) {
            return rows.get(row);
        }

        /**
         * Deletes the <code>MetaRecord</code> in the specified row.
         * @param row
         * @param row the row number, 0-based
         * @throws IndexOutOfBoundsException iff <code>row < 0</code>
         * or <code>row ></code> the maximum index  
         */
        public void delete(int row) {
            rows.remove(row);
            fireTableDataChanged();
        }

        /**
         * Deletes the specified <code>MetaRecord</code>.
         * @param record a <code>MetaRecord to be deleted</code>
         */
        public void delete(MetaRecord record) {
            rows.remove(record);
            fireTableDataChanged();
        }

        /**
         * Updates the <code>MetaRecord</code> at the specified index.
         * @param record a <code>MetaRecord to be updated</code>
         * @param index the index of the record
         * @throws IndexOutOfBoundsException iff <code>index < 0</code>
         * or <code>index ></code> the maximum index
         */
        public void update(MetaRecord record, int index) {
            rows.set(index, record);
            fireTableDataChanged();
        }

        public int getRowCount() {
            return rows.size();
        }

        public int getColumnCount() {
            int currentColumn = 0;
            for (int i = 0; i < metaScheme.getColumns().getSize(); i++) {
                DataColumn dataColumn = metaScheme.getColumns().get(i);
                if (!dataColumn.isHidden()) {
                    currentColumn++;
                }
            }
            return currentColumn;
        }

        public Object getValueAt(int rowIndex, int columnIndex) {
            MetaRecord record = rows.get(rowIndex);
            int virtualColumn = 0;
            for (int i = 0; i < metaScheme.getColumns().getSize(); i++) {
//            System.out.println(
//                metaScheme.getColumns().get(i).getName() + " " 
//                + metaScheme.getColumns().get(i).isHidden());
                while (metaScheme.getColumns().get(i).isHidden()) {
                    if (i >= metaScheme.getColumns().getSize()) {
                        throw new IndexOutOfBoundsException();
                    }

                    i++;
                }
                if (virtualColumn == columnIndex) {
                    return record.getValue(i);
                }
                virtualColumn++;
            }

            throw new IndexOutOfBoundsException();
        }
    }
}
