package org.kaiwitte.workbench.ui.view;

import org.kaiwitte.workbench.data.*;
import org.kaiwitte.workbench.ui.model.MetaSchemeTableModel;
import org.kaiwitte.workbench.ui.controller.*;
import org.kaiwitte.workbench.ui.UIConfig;
import org.kaiwitte.workbench.printing.PrintTemplateCreator;
import org.kaiwitte.workbench.tools.Tools;
import org.kaiwitte.tools.SelectAllListener;

import javax.swing.*;
import javax.swing.event.DocumentListener;
import javax.swing.event.DocumentEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import javax.swing.text.JTextComponent;
import java.awt.*;
import java.util.*;

import static java.lang.System.*;

import static java.awt.GridBagConstraints.*;

import net.tmorris.adt.sequence.Sequence;
import net.tmorris.adt.sequence.ArraySequenceImpl;

/**
 * A <code>JPanel</code> that allows the user to edit records.
 */
public final class EditRecordPanel extends JPanel implements MetaRecordCreator {
    private final MetaSchemeTableModel tableModel;
    private final MetaRecord originalRecord;
    private final UIConfig config;
    private int row = 0;
    private final ReturnKeyRemapper remapper =
        new ReturnKeyRemapper();

    private final EditableFocusTraversalPolicy policy =
        new EditableFocusTraversalPolicy();

    /**
     * Creates a new instance.
     * @param model the <code>MetaSchemeTableModel</code> that defines
     * the number and type of the editing fields.
     * @param index the index of the <code>MetaRecord</code> in the provided
     * <code>MetaSchemeTableModel</code>
     * @param config the <code>UIConfig</code>
     * @param record the <code>MetaRecord</code> to be edited 
     * @param dialog the <code>JDialog</code> to be closed after successful edit 
     */
    public EditRecordPanel(MetaSchemeTableModel model,
                           int index,
                           UIConfig config,
                           MetaRecord record,
                           JDialog dialog) {
        this.config = config;
        tableModel = model;
        this.originalRecord = record;
        setLayout(new GridBagLayout());
        addEditFields(model);

        // buttons
        GridBagConstraints c = new GridBagConstraints();

        final JComboBox combo = new JComboBox();
        final DataScheme dataScheme = model.getScheme().getDataSchemes().get(0);
        for (Iterator<String[]> it = dataScheme.printTemplates();
             it.hasNext();) {
            final String[] template = it.next();
            combo.addItem(Arrays.asList(template));
        }
        JButton print = new JButton(
            new PrintAction(config,
                            this,
                            new PrintTemplateCreator() {
                                public Sequence<String> templates() {
                                    if (combo.getSelectedItem() == null) {
                                        return null;
                                    }
                                    else {
                                        return
                                            ArraySequenceImpl.
                                                getArraySequence()
                                                .getSequence(
                                            ((java.util.List<String>)
                                            combo.getSelectedItem())
                                            .toArray(new String[0]));
                                    }
                                }
                            },
                            new Boolean(
                                config.getOption("close_after_edit_print")),
                            dialog
            ));


        JButton edit = new JButton(
            new EditRecordAction(config, model, index, this, dialog));
        edit.setMnemonic('e');
        print.setMnemonic('d');

        c.gridwidth = GridBagConstraints.RELATIVE;
        c.fill = GridBagConstraints.NONE;
        c.anchor = WEST;
        c.weightx = 0.0;
        c.gridy = row;
        add(new JLabel(config.getResource("print_template")), c);
        add(combo, c);
        c.gridy++;
        add(print, c);
        add(edit, c);
        remapper.remap(edit);
    }

    /**
     * Returns the <code>FocusTraversalPolicy</code> that is optimised
     * for navigating in this panel
     * @return the <code>FocusTraversalPolicy</code> 
     */
    public FocusTraversalPolicy getPolicy() {
        return policy;
    }

    private void addEditFields(MetaSchemeTableModel model) {
        GridBagConstraints c = new GridBagConstraints();

        // add id field
        c.fill = GridBagConstraints.HORIZONTAL;
        c.weightx = 0.5;
        c.gridx = 0;
        c.gridy = row;
        add(new JLabel("ID"), c);

        c.gridx = 1;
        c.gridy = row;
        add(new JLabel(Long.toString(originalRecord.getRecords().get(0).getID())),
            c);

        row++;

        for (int i = 0; i < model.getScheme().getDataSchemes().getSize(); i++) {
            DataScheme scheme = model.getScheme().getDataSchemes().get(i);
            Record record = originalRecord.getRecords().get(i);
            for (int j = 0; j < scheme.getColumns().getSize(); j++) {
                DataColumn column = scheme.getColumns().get(j);
                Object value = record.getValue(j);
                JLabel label = new JLabel(column.getName());
                ComponentCreator componentCreator =
                    CLASS_TO_COMPONENT.get(column.getType());

                /*
                * The resulting MetaScheme must have duplicate columns
                * for the associations (e. g. a product is bought by a
                * customer, which is again an entry in a different
                * database / scheme), but this panel must not.
                */
                if (column.getAssociatedDataScheme() == null) {
                    String columnID =
                        scheme.getName()
                        + "."
                        + column.getName();
                    JComponent component = componentCreator.createComponent(
                        column, columnID, value);

                    c.fill = GridBagConstraints.HORIZONTAL;
                    c.weightx = 0.5;
                    c.gridx = 0;
                    c.gridy = row;
                    int height = column.getTextAreaHeight();
                    if (height != 0) {
                        c.gridheight = height;
                        //c.fill = GridBagConstraints.BOTH;
                        //c.weighty = 4.0;
                    }
                    else {
                        c.gridheight = 1;
                        //c.fill = GridBagConstraints.HORIZONTAL;
                        //c.weighty = 1.0;
                    }
                    add(label, c);

                    c.gridx = 1;
                    c.gridy = row;
                    add(component, c);
                    remapper.remap(component);

                    policy.add(component);

                    row = row + c.gridheight;
                }
                else {
                    String columnID =
                        column.getAssociatedDataScheme().getName()
                        + "."
                        + column.getNameInAssociatedDataScheme();
                    componentCreator.createComponent(
                        column, columnID, value);
                }
            }
        }
    }

    /**
     * Creates and returns the <code>MetaRecord</code> that results from
     * the changes made in this dialog.
     * @return the new <code>MetaRecord</code>
     */
    public MetaRecord createMetaRecord() {
        Sequence<Record> originalRecords = originalRecord.getRecords();
        MetaScheme metaScheme = tableModel.getScheme();
        java.util.List<Record> records = new java.util.ArrayList<Record>();
        int i = 0;
        int j = 0;
        for (DataScheme scheme : metaScheme.getDataSchemes()) {
            java.util.List values = new ArrayList();

            for (DataColumn column : scheme.getColumns()) {
                values.add(valueCreators.get(i++).getValue());
            }
            Record record = new MutableRecordImpl(
                originalRecords.get(j++).getID(), scheme, values.toArray());
            records.add(record);
        }
        return new MetaRecordImpl(records.toArray(new Record[0]));
    }

    /**
     * Returns the <code>MetaSchemeTableModel</code> that belongs to
     * this panel.
     * @return the <code>MetaSchemeTableModel</code>
     * @see #EditRecordPanel
     */
    public MetaSchemeTableModel getTableModel() {
        return tableModel;
    }

    /**
     * Creates a JComponent for an ???.
     * Example: Create a <code>JCheckBox</code> if the associated
     * <code>Option</code>'s type is <code>Boolean.class</code>.
     */
    private static interface ComponentCreator {
        public JComponent createComponent(DataColumn column, String key,
                                          Object value);
    }
        /**
         * A <code>Map</code> to associate a <code>Class</code> with a
         * {@link org.kaiwitte.workbench.ui.view.EditRecordPanel.ComponentCreator} that creates the <code>JComponent</code>
         * that is best to represent the value of an instance of this 
         * <code>Class</code>.
         * Example: A <code>Boolean.class</code> is mapped to a
         * {@link org.kaiwitte.workbench.ui.view.EditRecordPanel.ComponentCreator} that creates a <code>JCheckBox</code>.
         */
        private static final Map<Class,ComponentCreator> CLASS_TO_COMPONENT =
            new HashMap<Class,ComponentCreator>();
        {
            CLASS_TO_COMPONENT.put(String.class, new ComponentCreator() {
                public JComponent createComponent(DataColumn column,
                                                  String key, Object value) {
                    JTextComponent text;
                    if (column.getTextAreaHeight() == 0) {
                        text = new JTextField(15);
                    }
                    else {
                        JTextArea text_ =
                            new JTextArea(column.getTextAreaHeight(),
                                          15);
                        text_.setLineWrap(true);
                        text = text_;
                        text.setBorder(new JTextField().getBorder());
                        /*text.setPreferredSize(
                            new Dimension(15, 40)
                        );
                        text.setSize(15, 40);*/
                    }

                    text.addFocusListener(new SelectAllListener());
                    assert value instanceof String;
                    text.setText((String) value);
                    //text.addFocusListener(new SelectAllListener());
                    if (reusableValueCreators.get(key) == null) {
                        reusableValueCreators.put(key,
                                                  new StringValueCreator(text));
                    }
                    else {
                        ValueCreator valueCreator = new StringValueCreator(text);
                        valueCreators.set(
                            valueCreators.indexOf(
                                reusableValueCreators.get(key)),
                            valueCreator
                        );
                        reusableValueCreators.put(key,
                                                  valueCreator);
                    }
                    valueCreators.add(reusableValueCreators.get(key));
                    return text;
                }
            });
            CLASS_TO_COMPONENT.put(String[].class, new ComponentCreator() {
                public JComponent createComponent(DataColumn column,
                                                  String key, Object value) {
                    JComboBox text = new JComboBox(
                        Tools.sequenceToArray(column.getChoices(), 
                                              String[].class));
                    assert value instanceof String;
                    text.setSelectedItem(value);
                    //text.addFocusListener(new SelectAllListener());
                    if (reusableValueCreators.get(key) == null) {
                        reusableValueCreators.put(key,
                                                  new ChoiceValueCreator(text));
                    }
                    else {
                        ValueCreator valueCreator = new ChoiceValueCreator(text);
                        valueCreators.set(
                            valueCreators.indexOf(
                                reusableValueCreators.get(key)),
                            valueCreator
                        );
                        reusableValueCreators.put(key,
                                                  valueCreator);
                    }
                    valueCreators.add(reusableValueCreators.get(key));
                    return text;
                }
            });
            CLASS_TO_COMPONENT.put(Integer.class, new ComponentCreator() {
                public JComponent createComponent(DataColumn column,
                                                  String key, Object value) {
                    JSpinner spinner = new JSpinner();
                    assert value instanceof Integer;
                    spinner.setValue(value);
                    //spinner.setValue(new Integer(o.createCondition()));
                    if (reusableValueCreators.get(key) == null) {
                        reusableValueCreators.put(key,
                                                  new IntValueCreator(spinner));
                    }
                    else {
                        ValueCreator valueCreator = new IntValueCreator(spinner);
                        valueCreators.set(
                            valueCreators.indexOf(
                                reusableValueCreators.get(key)),
                            valueCreator
                        );
                        reusableValueCreators.put(key,
                                                  valueCreator);
                    }
                    valueCreators.add(reusableValueCreators.get(key));
                    return spinner;
                }
            });
            CLASS_TO_COMPONENT.put(java.util.Date.class, new ComponentCreator() {
                public JComponent createComponent(DataColumn column,
                                                  String key, Object value) {
                    JSpinner spinner = new JSpinner();
                    spinner.setModel(new SpinnerDateModel());
                    spinner.setEditor(new JSpinner.DateEditor(spinner,
                                                              "dd.MM.yyyy"));
                    assert value instanceof java.util.Date;
                    spinner.setValue(value);
                    //spinner.setValue(new Integer(o.createCondition()));
                    if (reusableValueCreators.get(key) == null) {
                        reusableValueCreators.put(key,
                                                  new DateValueCreator(spinner));
                    }
                    else {
                        ValueCreator valueCreator = new DateValueCreator(spinner);
                        valueCreators.set(
                            valueCreators.indexOf(
                                reusableValueCreators.get(key)),
                            valueCreator
                        );
                        reusableValueCreators.put(key,
                                                  valueCreator);
                    }
                    valueCreators.add(reusableValueCreators.get(key));
                    return spinner;
                }
            });
            CLASS_TO_COMPONENT.put(Boolean.class, new ComponentCreator() {
                public JComponent createComponent(DataColumn column,
                                                  String key, Object value) {
                    JCheckBox spinner = new JCheckBox();
                    assert value instanceof Boolean;
                    spinner.setSelected((Boolean) value);
                    //spinner.setValue(new Integer(o.createCondition()));
                    if (reusableValueCreators.get(key) == null) {
                        reusableValueCreators.put(key,
                                                  new BooleanValueCreator(spinner));
                    }
                    else {
                        ValueCreator valueCreator = new BooleanValueCreator(spinner);
                        valueCreators.set(
                            valueCreators.indexOf(
                                reusableValueCreators.get(key)),
                            valueCreator
                        );
                        reusableValueCreators.put(key,
                                                  valueCreator);
                    }
                    valueCreators.add(reusableValueCreators.get(key));
                    return spinner;
                }
            });
//            CLASS_TO_COMPONENT.put(Boolean.class, new ComponentCreator() {
//                public JComponent createComponent(DataColumn column,
//                                                  ButtonModel model) {
//                    JCheckBox button = new JCheckBox();
//                    //button.setSelected(Boolean.parseBoolean(o.createCondition()));
//                    conditionCreators.put(column, new BooleanConditionCreator(button,
//                                                                          column));
//                    return button;
//                }
//            });
        }
    /**
     * Reads a value from a swing model object.
     * Example: If the associated model is a <code>Document</code> it reads
     * the containing text.
     */
    private static interface ValueCreator {
        public Object getValue();
    }

    /**
     * A {@link org.kaiwitte.workbench.ui.view.EditRecordPanel.ValueCreator} for a <code>JTextComponent</code>.
     */
    private static class StringValueCreator implements ValueCreator {
        private JTextComponent text;

        public StringValueCreator(JTextComponent text) {
            this.text = text;
        }

        public Object getValue() {
            return text.getText();
        }
    }

    /**
     * A {@link org.kaiwitte.workbench.ui.view.EditRecordPanel.ValueCreator} for a <code>JTextComponent</code>.
     */
    private static class ChoiceValueCreator implements ValueCreator {
        private JComboBox text;

        public ChoiceValueCreator(JComboBox text) {
            this.text = text;
        }

        public Object getValue() {
            return text.getSelectedItem();
        }
    }

    /**
     * A {@link org.kaiwitte.workbench.ui.view.EditRecordPanel.ValueCreator} for a <code>JSpinner</code>.
     */
    private static class IntValueCreator implements ValueCreator {
        private JSpinner text;

        public IntValueCreator(JSpinner text) {
            this.text = text;
        }

        public Object getValue() {
            return text.getValue();
        }
    }

    /**
     * A {@link org.kaiwitte.workbench.ui.view.EditRecordPanel.ValueCreator} for a <code>JSpinner</code>.
     */
    private static class DateValueCreator implements ValueCreator {
        private JSpinner text;

        public DateValueCreator(JSpinner text) {
            this.text = text;
        }

        public Object getValue() {
            java.util.Date date =
                ((SpinnerDateModel) text.getModel()).getDate();
            return date;
        }
    }

    /**
     * A {@link org.kaiwitte.workbench.ui.view.EditRecordPanel.ValueCreator} for a <code>JCheckBox</code>.
     */
    private static class BooleanValueCreator implements ValueCreator {
        private JCheckBox text;

        public BooleanValueCreator(JCheckBox text) {
            this.text = text;
        }

        public Object getValue() {
            return text.isSelected();
        }
    }

    /**
     * A map to associate a single {@link org.kaiwitte.workbench.data.DataColumn} with a
     * {@link org.kaiwitte.workbench.ui.view.EditRecordPanel.ValueCreator} that reads the new value for this option
     * from a <code>JComponent</code>.
     * Example: The {@link org.kaiwitte.workbench.data.DataColumn} has the type <code>Boolean.class</code>.
     * Then it has an associated <code>JCheckbox</code> in the 
     * <code>JDialog</code>. This <code>Map</code> contains the
     * {@link org.kaiwitte.workbench.ui.view.EditRecordPanel.ValueCreator} that reads the status of the <code>JCheckBox</code>
     * (checked or not).
     */
    private final java.util.List<ValueCreator> valueCreators =
        new ArrayList<ValueCreator>();

    private final Map<String,ValueCreator> reusableValueCreators =
        new HashMap<String,ValueCreator>();
}
