package org.kaiwitte.workbench.ui.view;

import org.kaiwitte.workbench.ui.controller.*;
import org.kaiwitte.workbench.ui.UIConfig;
import org.kaiwitte.joptions.OptionManager;
import org.kaiwitte.tools.SwingTools;

import javax.swing.*;
import java.awt.event.KeyEvent;
import java.lang.reflect.InvocationTargetException;

/**
 * The main frame.
 */
public final class MainView extends JFrame {
    private final JTabbedPane tabbedPane = new JTabbedPane();
    private volatile int selection;
    private OptionManager optionManager;
    private final UIConfig config;

    /**
     * Creates a new instance.
     * @param config the <code>UIConfig</code>
     */
    public MainView(UIConfig config) {
        super("Datenkai");
        this.config = config;

        //setSize(400, 400);
        setContentPane(tabbedPane);
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        JMenuBar menuBar = buildMenu(config);

        setJMenuBar(menuBar);
    }

    private JMenuBar buildMenu(UIConfig config) {
        JMenuBar menuBar = new JMenuBar();
        JMenu file = new JMenu(config.getResource("file"));
        file.setMnemonic(config.getResource("file_m").charAt(0));

        ShowOptionsAction optionsAction = new ShowOptionsAction(config);
        JMenuItem optionsMenu = new JMenuItem(optionsAction);
        optionsMenu.setMnemonic('o');
        file.add(optionsMenu);

        Action exitAction = new ExitAction(config);
        JMenuItem exitMenu = new JMenuItem(exitAction);
        exitMenu.setMnemonic(config.getResource("exit_m").charAt(0));
        file.add(exitMenu);

        JMenu database = new JMenu(config.getResource("database"));
        database.setMnemonic(config.getResource("database_m").charAt(0));

        Action reconnectAction = new ReconnectAction(config);
        JMenuItem reconnectMenu = new JMenuItem(reconnectAction);
        reconnectMenu.setMnemonic(config.getResource("reconnect_m").charAt(0));
        database.add(reconnectMenu);

        JMenu help = new JMenu(config.getResource("help"));
        help.setMnemonic(config.getResource("help_m").charAt(0));

        Action showHelpAction = new ShowHelpAction(config);
        JMenuItem showHelp = new JMenuItem(showHelpAction);
        showHelp.setMnemonic(config.getResource("help_m").charAt(0));
        help.add(showHelp);
        // F1 shows help:
        KeyStroke escape = KeyStroke.getKeyStroke(KeyEvent.VK_F1, 0, false);
        getRootPane().getInputMap(
            JComponent.WHEN_IN_FOCUSED_WINDOW).put(escape, "HELP");
        getRootPane().getActionMap().put("HELP", showHelpAction);

        Action showInfoAction = new InfoAction(config);
        JMenuItem showInfo = new JMenuItem(showInfoAction);
        showInfo.setMnemonic(config.getResource("info_m").charAt(0));
        help.add(showInfo);

        JMenu developer = new JMenu(config.getResource("developer"));
        developer.setMnemonic(config.getResource("developer_m").charAt(0));

        Action showStdOutAction = new ShowTextAction(config);
        JMenuItem showdeveloper = new JMenuItem(showStdOutAction);
        showdeveloper.setMnemonic(config.getResource("show_out_m").charAt(0));
        developer.add(showdeveloper);

        Action showHelpDevAction = new ShowDeveloperHelpAction(config);
        JMenuItem showHelpDev = new JMenuItem(showHelpDevAction);
        showHelpDev.setMnemonic(
            config.getResource("help_for_developers_m").charAt(0));
        developer.add(showHelpDev);

        Action showFeedback = new ShowFeedbackDialogAction(config);
        JMenuItem showFeedback_m = new JMenuItem(showFeedback);
        showFeedback_m.setMnemonic(
            config.getResource("feedback_m").charAt(0));
        developer.add(showFeedback_m);

        menuBar.add(file);
        menuBar.add(database);
        menuBar.add(developer);
        menuBar.add(help);
        return menuBar;
    }

    /**
     * Adds a <code>JPanel</code> as a tab.
     * @param title the title to be displayed in the tab
     * @param tab the new panel
     */
    public void addTab(String title, JPanel tab) {
        tabbedPane.add(title, tab);
    }

    /**
     * Displays an information message.
     * @param message the message
     */
    public void info(final String message) {
        Runnable runnable = new Runnable() {
            public void run() {
                JOptionPane.showMessageDialog(
                    MainView.this,
                    message,
                    "information",
                    JOptionPane.INFORMATION_MESSAGE
                );
            }
        };
        SwingUtilities.invokeLater(runnable);
    }

    /**
     * Displays an error message.
     * @param message the message
     * @param t the <code>Throwable</code> that caused the error
     */
    public void error(final String message, final Throwable t) {
        Runnable runnable = new Runnable() {
            public void run() {
                SwingTools.exceptionDialog(MainView.this, message, t);
            }
        };
        SwingUtilities.invokeLater(runnable);
    }

    /**
     * Displays an error message.
     * @param message the message
     */
    public void error(final String message) {
        Runnable runnable = new Runnable() {
            public void run() {
                JOptionPane.showMessageDialog(
                    MainView.this,
                    message,
                    "error",
                    JOptionPane.ERROR_MESSAGE
                );
            }
        };
        SwingUtilities.invokeLater(runnable);
    }

    /**
     * Displays a confirmation dialog.
     * @param message the message for the dialog
     * @return <code>true</code> iff the user approved
     */
    public boolean confirm(final String message) {
        selection = JOptionPane.NO_OPTION;
        Runnable runnable = new Runnable() {
            public void run() {
                selection = JOptionPane.showConfirmDialog(
                    MainView.this,
                    message,
                    "confirm",
                    JOptionPane.YES_NO_OPTION,
                    JOptionPane.QUESTION_MESSAGE
                );
            }
        };

        if (!SwingUtilities.isEventDispatchThread()) {
            try {
                SwingUtilities.invokeAndWait(runnable);
            } catch (InterruptedException e) {
                config.getMainModel().logException(e);
                throw new AssertionError(e);
            } catch (InvocationTargetException e) {
                config.getMainModel().logException(e);
                throw new AssertionError(e);
            }
        }
        else {
            runnable.run();
        }

        return selection == JOptionPane.YES_OPTION;
    }

    /**
     * Displays the option dialog.
     */
    public void showOptions() {
        optionManager.showDialog(this);
    }

    /**
     * Sets the <code>OptionManager</code> to be used for the
     * {@link #showOptions()}.
     * @param optionManager the <code>OptionManager</code> to be used
     */
    public void setOptionManager(OptionManager optionManager) {
        this.optionManager = optionManager;
    }
}
