package org.kaiwitte.workbench.ui.view;

import org.kaiwitte.workbench.data.*;
import org.kaiwitte.workbench.ui.model.MetaSchemeTableModel;
import org.kaiwitte.workbench.ui.controller.*;
import org.kaiwitte.workbench.ui.UIConfig;
import org.kaiwitte.workbench.printing.PrintTemplateCreator;
import org.kaiwitte.workbench.tools.Tools;
import org.kaiwitte.tools.SelectAllListener;

import javax.swing.*;
import javax.swing.event.DocumentListener;
import javax.swing.event.DocumentEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import javax.swing.text.JTextComponent;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.util.*;
import java.util.List;

import static java.lang.System.*;

import static java.awt.GridBagConstraints.*;

import net.tmorris.adt.sequence.Sequence;
import net.tmorris.adt.sequence.ArraySequenceImpl;

/**
 * A <code>JPanel</code> that allows the user to create a new record.
 */
public final class NewRecordPanel extends JPanel implements RecordReceiver,
        MetaRecordCreator {
    private final MetaSchemeTableModel tableModel;
    private final GridBagConstraints c = new GridBagConstraints();
    private final UIConfig config;
    private final ReturnKeyRemapper remapper =
        new ReturnKeyRemapper();

    private final EditableFocusTraversalPolicy policy =
        new EditableFocusTraversalPolicy();
    private final Map<DataColumn,JComponent> columnToComponent =
        new HashMap<DataColumn,JComponent>();
    private final StoreRecordAction storeRecordAction;
    private List<JComponent> requiredFields = new ArrayList<JComponent>();
    private JButton storeButton;

    /**
     * Creates a new instance.
     * @param model the <code>MetaSchemeTableModel</code> that defines
     * the number and type of the editing fields.
     * @param config the <code>UIConfig</code>
     * @param dialog the surrounding <code>JDialog</code> to be closed
     */
    public NewRecordPanel(MetaSchemeTableModel model,
                          UIConfig config,
                          JDialog dialog) {
        this.config = config;
        tableModel = model;
        c.anchor = NORTHWEST;
        setLayout(new GridBagLayout());
        addSearchFields(model);
        //c.gridy = 0;
        //add(new JScrollPane(new JTextArea("hello")), c);
        //c.gridy++;

        // print templates
        //c.anchor = WEST;
        c.gridwidth=1;
        add(new JLabel(config.getResource("print_template")), c);
        c.gridx++;
        final JComboBox combo = new JComboBox();
        final DataScheme dataScheme = model.getScheme().getDataSchemes().get(0);
        for (Iterator<String[]> it = dataScheme.printTemplates();
             it.hasNext();) {
            final String[] template = it.next();
            combo.addItem(Arrays.asList(template));
        }
        add(combo, c);
        c.gridy++;

        // buttons
        storeRecordAction = new StoreRecordAction(config, this, dialog);
        storeButton = new JButton(storeRecordAction);
        c.gridwidth = 1;
        //c.fill = GridBagConstraints.NONE;
        //c.anchor = WEST;
        //c.weightx = 0.0;
        c.gridx = 0;
        storeButton.setMnemonic('s');
        policy.add(storeButton);
        remapper.remap(storeButton);
        add(storeButton, c);
        c.gridx = 1;
        //c.gridy++;

        JButton print = new JButton(
            new PrintAction(config,
                            this,
                            new PrintTemplateCreator() {
                                public Sequence<String> templates() {
                                    if (combo.getSelectedItem() == null) {
                                        return null;
                                    }
                                    else {
                                        return ArraySequenceImpl
                                            .getArraySequence()
                                            .getSequence(((java.util.List<String>)
                                            combo.getSelectedItem())
                                            .toArray(new String[0]));
                                    }
                                }
                            },
                            new Boolean(
                                config.getOption("close_after_save_print")),
                            dialog
            ));
        c.gridwidth = GridBagConstraints.RELATIVE;
        c.fill = GridBagConstraints.NONE;
        c.anchor = WEST;
        c.weightx = 0.0;
        c.gridx = 1;
        print.setMnemonic('p');
        policy.add(print);
        remapper.remap(print);
        add(print, c);

        // compound print & store button?
        if (new Boolean(config.getOption("print_and_store"))
                && dataScheme.printTemplates().hasNext()) {
            CompoundAction storeAndPrintAction =
                new CompoundAction(storeButton.getAction(), print.getAction());
            storeAndPrintAction.putValue(Action.NAME,
                                         config.getResource("store_and_print"));
            storeAndPrintAction.putValue(Action.SMALL_ICON, new ImageIcon(
                getClass().getResource("/res/Save24.gif")));
            storeButton.setAction(storeAndPrintAction);
        }
        storeButton.getAction().setEnabled(false);

        c.gridy++;

        //c.gridwidth = 2;
        c.gridx = 0;
    }

    /**
     * Returns the <code>FocusTraversalPolicy</code> that is optimised
     * for navigating in this panel
     * @return the <code>FocusTraversalPolicy</code> 
     */
    public FocusTraversalPolicy getPolicy() {
        return policy;
    }

    private void addSearchFields(MetaSchemeTableModel model) {
        List<JComponent> searchFieldPanels = new LinkedList<JComponent>();
        for (DataScheme scheme : model.getScheme().getDataSchemes()) {
            JPanel searchFieldPanel = new JPanel(new GridBagLayout());
            GridBagConstraints c = new GridBagConstraints(); // hides global c!
            c.gridy = 0;
            c.gridx = 0;

            for (DataColumn column : scheme.getColumns()) {
                JLabel label = new JLabel(column.getName());
                if (column.isRequiredForUpdate()) {
                    //out.println("required: " + column.getName());
                    label.setText(label.getText()+"*");
                }
                ComponentCreator componentCreator =
                    CLASS_TO_COMPONENT.get(column.getType());

                /*
                * The resulting MetaScheme must have duplicate columns
                * for the associations (e. g. a product is bought by a
                * customer, which is again an entry in a different
                * database / scheme), but this panel must not.
                */
                if (column.getAssociatedDataScheme() == null) {
                    String columnID =
                        scheme.getName()
                        + "."
                        + column.getName();
                    String qualifiedName = columnID;
                    JComponent component = componentCreator.createComponent(
                        column, columnID);
                    columnToComponent.put(column, component);
                    if (column.isRequiredForUpdate()) {
                        requiredFields.add(component);
                    }

                    /*
                    * small column ?
                    * have two things in a row : have one thing in a row 
                    */
//                    if (column.isSmall()) {
//                        c.gridwidth = 1;
//                    }
//                    else {
//                        c.gridwidth = 1;
//                    }
                    c.gridwidth = 1;

                    c.fill = GridBagConstraints.HORIZONTAL;
                    c.weightx = 0.5;
                    int height = column.getTextAreaHeight();
                    if (height != 0) {
                        c.gridheight = height;
                        //c.fill = GridBagConstraints.BOTH;
                        //c.weighty = 4.0;
                    }
                    else {
                        c.gridheight = 1;
                        //c.fill = GridBagConstraints.HORIZONTAL;
                        //c.weighty = 1.0;
                    }
                    /*
                     * Special last minute customer requirement:
                     * Under circumstances that are just as complex as the
                     * following condition, the field is not to be displayed. 
                     */
                    boolean isMainColumn = 
                        model.getScheme().getDataSchemes().get(0) == scheme;
                    if (!(column.isIndirectlyHidden() && !isMainColumn)) {
                        out.println("NOT hiding: " + column.getName());
                        searchFieldPanel.add(label, c);

                        c.gridx++;
                        if (!column.isSmall()) {
                            c.gridwidth = 3;
                        }
                        searchFieldPanel.add(component, c);

                        /*
                        * small column ?
                        * have two things in a row : have one thing in a row 
                        */
                        if (column.isSmall() && c.gridx < 3) {
                            c.gridx++;
                        }
                        else {
                            c.gridx = 0;
                            c.gridy += c.gridheight;
                        }
                    }
                    else {
                        out.println("hiding: " + column.getName());
                        c.gridy += c.gridheight;
                    }


                    policy.add(component);
                    remapper.remap(component);

                    addTemplateButtons(searchFieldPanel, c, column, component);
                }
                else {
                    String columnID =
                        column.getAssociatedDataScheme().getName()
                        + "."
                        + column.getNameInAssociatedDataScheme();
                    String qualifiedName =
                        scheme.getName()
                        + "."
                        + column.getName();
                    JComponent component = componentCreator.createComponent(
                        column, columnID);
                    columnToComponent.put(column, component);
                }
            }
            searchFieldPanels.add(0, searchFieldPanel);
        }
        c.gridwidth=2;
        c.gridx = 0;
        c.gridy = 0;
        c.fill=HORIZONTAL;
//        JFrame tf = new JFrame();
//        tf.add(searchFieldPanels.get(0));
//        tf.pack();
//        tf.setVisible(true);
        for (JComponent panel : searchFieldPanels) {
            add(panel, c);
            //add(new JTextArea("test"), c);
            c.gridy++;
        }
        c.gridwidth=1;
    }

    private void addTemplateButtons(JComponent container,
                                    GridBagConstraints c,
                                    DataColumn column,
                                    final JComponent component) {
        java.util.List<JButton> buttons = new ArrayList<JButton>();

        for (char templateNumber = 'a';
             templateNumber <= 'd';
             templateNumber++) {
            final String templateName = "Fieldname" + templateNumber;
            String templateOption =
                config.getOption(templateName);
            if (column.getName().equals(templateOption)) {

                for (int i = 1; i <= 10; i++) {
                    final String fieldName =
                        config.getOption("Default" + i + templateNumber);
                    if (fieldName.length() != 0) {
                        Action applyTemplateAction =
                            new AbstractAction(fieldName) {
                                public void actionPerformed(ActionEvent e) {
                                    if (component instanceof JTextComponent) {
                                        JTextComponent text =
                                            ((JTextComponent) component);
                                        text.setText(
                                            (text.getText() + fieldName + " "));
                                    }
                                }
                            };
                        buttons.add(new JButton(applyTemplateAction));
                    }
                }
            }
        }

        if (buttons.size() == 0) {
            return;
        }

        JPanel templatePanel = new JPanel();
        templatePanel.setLayout(new BoxLayout(templatePanel, BoxLayout.X_AXIS));
        for (JButton button : buttons) {
            templatePanel.add(button);
        }
        c.gridheight = 1;
        c.gridx = 0;
        container.add(new JLabel(), c);
        c.gridx = 1;
        container.add(templatePanel, c);
        c.gridy++;
        c.gridx = 0;

    }

    /**
     * Creates and returns the <code>MetaRecord</code> that results from
     * the changes made in this dialog.
     * @return the new <code>MetaRecord</code>
     */
    public MetaRecord createMetaRecord() {
        MetaScheme metaScheme = tableModel.getScheme();
        java.util.List values = new ArrayList();
        for (ValueCreator vc : valueCreators) {
            values.add(vc.getValue());
        }
        MutableMetaRecord ret =
            new MutableMetaRecordImpl(metaScheme, values.toArray());

        MetaRecord insertedRecord = storeRecordAction.getInsertedRecord();

        return insertedRecord != null ? insertedRecord : ret;
    }

    /**
     * Returns the <code>MetaSchemeTableModel</code>.
     * @return the <code>MetaSchemeTableModel</code>
     * @see NewRecordPanel#NewRecordPanel
     */
    public MetaSchemeTableModel getTableModel() {
        return tableModel;
    }

    /**
     * Adds a <code>DocumentListener</code> to all visible text fields.
     * @param documentListener a <code>DocumentListener</code> to be added
     */
    public void addDocumentListener(DocumentListener documentListener) {
        for (DataColumn column : tableModel.getScheme().getColumns()) {
            JComponent component = columnToComponent.get(column);
            if (component instanceof JTextComponent) {
                ((JTextComponent) component).getDocument().
                    addDocumentListener(
                            documentListener);
            }
        }
    }

    /**
     * Sends a <code>MetaRecord</code> to this panel. Its content is displayed
     * in the panel. Used for quicksearch-and-insert-features.
     * @param metaRecord the new values
     * @param outerScheme the <code>MetaScheme</code> of the provided
     * <code>MetaRecord</code>
     */
    public void receiveMetaRecord(MetaRecord metaRecord,
                                  MetaScheme outerScheme) {
        Sequence<DataColumn> innerColumns = tableModel.getScheme().getColumns();
        Sequence<DataColumn> outerColumns = outerScheme.getColumns();

        for (int i = 0;
             i < innerColumns.getSize() && i < outerColumns.getSize();
             i++) {
            final DataColumn innerColumn = innerColumns.get(i);
            final DataColumn outerColumn = outerColumns.get(i);
            if (innerColumn.getType().equals(outerColumn.getType())
                && innerColumn.getAssociatedDataScheme() == null
                && outerColumn.getAssociatedDataScheme() == null) {
                JComponent component = columnToComponent.get(innerColumn);
                ValueApplier valueApplier = valueAppliers.get(
                    innerColumn.getType());
                valueApplier.applyValue(component, metaRecord.getValue(i));

            }
        }
    }

    /**
     * Sends a <code>Record</code> to this panel. Its content is displayed
     * in the panel. Used for quicksearch-and-insert-features.
     * @param r the new values
     */
    public void receiveRecord(Record r) {
        Sequence<DataColumn> columns = r.getScheme().getColumns();
        for (int i = 0; i < columns.getSize(); i++) {
            DataColumn column = columns.get(i);
            if (column.getAssociatedDataScheme() != null) {
                // don't allow specification of external fields
                continue;
            }

            JComponent component = columnToComponent.get(column);
            ValueApplier valueApplier = valueAppliers.get(column.getType());
            valueApplier.applyValue(component, r.getValue(i));
        }
    }

    private static interface ValueApplier {
        public void applyValue(JComponent c, Object value);
    }

    private static final class StringValueApplier implements ValueApplier {
        public void applyValue(JComponent c, Object value) {
            ((JTextComponent) c).setText((String) value);
        }
    }

    private static final class SpinnerValueApplier implements ValueApplier {
        public void applyValue(JComponent c, Object value) {
            ((JSpinner) c).setValue(value);
        }
    }

    private static final class ChoiceValueApplier implements ValueApplier {
        public void applyValue(JComponent c, Object value) {
            ((JComboBox) c).setSelectedItem(value);
        }
    }

    private static final class BooleanValueApplier implements ValueApplier {
        public void applyValue(JComponent c, Object value) {
            ((JCheckBox) c).setSelected((Boolean) value);
        }
    }

    private final Map<Class,ValueApplier> valueAppliers =
        new HashMap<Class,ValueApplier>();
    {
        valueAppliers.put(String.class, new StringValueApplier());
        valueAppliers.put(String[].class, new ChoiceValueApplier());
        valueAppliers.put(java.util.Date.class, new SpinnerValueApplier());
        valueAppliers.put(Integer.class, new SpinnerValueApplier());
        valueAppliers.put(Boolean.class, new BooleanValueApplier());
    }

    /**
     * Creates a JComponent for an ???.
     * Example: Create a <code>JCheckBox</code> if the associated
     * <code>Option</code>'s type is <code>Boolean.class</code>.
     */
    private static interface ComponentCreator {
        public JComponent createComponent(DataColumn column, String key);
    }
        /**
         * A <code>Map</code> to associate a <code>Class</code> with a
         * {@link org.kaiwitte.workbench.ui.view.NewRecordPanel.ComponentCreator} that creates the <code>JComponent</code>
         * that is best to represent the value of an instance of this 
         * <code>Class</code>.
         * Example: A <code>Boolean.class</code> is mapped to a
         * {@link org.kaiwitte.workbench.ui.view.NewRecordPanel.ComponentCreator} that creates a <code>JCheckBox</code>.
         */
        private static final Map<Class,ComponentCreator> CLASS_TO_COMPONENT =
            new HashMap<Class,ComponentCreator>();
    private List<String> requiredEmptyFields = new ArrayList<String>();

    {
        CLASS_TO_COMPONENT.put(String.class, new ComponentCreator() {
            public JComponent createComponent(final DataColumn column,
                                              String key) {
                JTextComponent text;
                if (column.getTextAreaHeight() == 0) {
                    text = new JTextField(15);
                }
                else {
                    JTextArea text_ =
                        new JTextArea(column.getTextAreaHeight(),
                                      15);
                    text_.setLineWrap(true);
                    text = text_;
                    text.setBorder(new JTextField().getBorder());
                    /*text.setPreferredSize(
                       new Dimension(15, 40)
                   );
                   text.setSize(15, 40);*/
                }
                text.addFocusListener(new SelectAllListener());

                if (column.isRequiredForUpdate()) {
                    // field is mandatory - watch when it is filled in
                    requiredEmptyFields.add(column.getName());
                    text.getDocument().addDocumentListener(
                        new DocumentListener() {
                            public void insertUpdate(DocumentEvent e) {
                                requiredEmptyFields.remove(column.getName());
                                if (requiredEmptyFields.isEmpty()) {
                                    // all mandatory fields are filled in
                                    storeButton.getAction().setEnabled(true);
                                }
                            }

                            public void removeUpdate(DocumentEvent e) { }

                            public void changedUpdate(DocumentEvent e) { }
                        });
                }
                //text.addFocusListener(new SelectAllListener());
                if (reusableValueCreators.get(key) == null) {
                    reusableValueCreators.put(key,
                                              new StringValueCreator(text));
                }
                else {
                    ValueCreator valueCreator = new StringValueCreator(text);
                    valueCreators.set(
                        valueCreators.indexOf(
                            reusableValueCreators.get(key)),
                        valueCreator
                    );
                    reusableValueCreators.put(key,
                                              valueCreator);
                }
                valueCreators.add(reusableValueCreators.get(key));
                return text;
            }
        });
        CLASS_TO_COMPONENT.put(String[].class, new ComponentCreator() {
            public JComponent createComponent(DataColumn column,
                                              String key) {
                JComboBox text = new JComboBox(
                    Tools.sequenceToArray(column.getChoices(), String[].class));
                //text.addFocusListener(new SelectAllListener());
                if (reusableValueCreators.get(key) == null) {
                    reusableValueCreators.put(key,
                                              new ChoiceValueCreator(text));
                }
                else {
                    ValueCreator valueCreator = new ChoiceValueCreator(text);
                    valueCreators.set(
                        valueCreators.indexOf(
                            reusableValueCreators.get(key)),
                        valueCreator
                    );
                    reusableValueCreators.put(key,
                                              valueCreator);
                }
                valueCreators.add(reusableValueCreators.get(key));
                return text;
            }
        });
        CLASS_TO_COMPONENT.put(Integer.class, new ComponentCreator() {
            public JComponent createComponent(DataColumn column,
                                              String key) {
                JSpinner spinner = new JSpinner();
                //spinner.setValue(new Integer(o.createCondition()));
                if (reusableValueCreators.get(key) == null) {
                    reusableValueCreators.put(key,
                                              new IntValueCreator(spinner));
                }
                else {
                    ValueCreator valueCreator = new IntValueCreator(spinner);
                    valueCreators.set(
                        valueCreators.indexOf(
                            reusableValueCreators.get(key)),
                        valueCreator
                    );
                    reusableValueCreators.put(key,
                                              valueCreator);
                }
                valueCreators.add(reusableValueCreators.get(key));
                return spinner;
            }
        });
        CLASS_TO_COMPONENT.put(java.util.Date.class, new ComponentCreator() {
            public JComponent createComponent(DataColumn column,
                                              String key) {
                JSpinner spinner = new JSpinner();
                spinner.setModel(new SpinnerDateModel());
                spinner.setEditor(new JSpinner.DateEditor(spinner,
                                                          "dd.MM.yyyy"));

                //spinner.setValue(new Integer(o.createCondition()));
                if (reusableValueCreators.get(key) == null) {
                    reusableValueCreators.put(key,
                                              new DateValueCreator(spinner));
                }
                else {
                    ValueCreator valueCreator = new DateValueCreator(spinner);
                    valueCreators.set(
                        valueCreators.indexOf(
                            reusableValueCreators.get(key)),
                        valueCreator
                    );
                    reusableValueCreators.put(key,
                                              valueCreator);
                }
                valueCreators.add(reusableValueCreators.get(key));
                return spinner;
            }
        });
        CLASS_TO_COMPONENT.put(Boolean.class, new ComponentCreator() {
            public JComponent createComponent(DataColumn column,
                                              String key) {
                JCheckBox spinner = new JCheckBox();
                //spinner.setValue(new Integer(o.createCondition()));
                if (reusableValueCreators.get(key) == null) {
                    reusableValueCreators.put(key,
                                              new BooleanValueCreator(spinner));
                }
                else {
                    ValueCreator valueCreator = new BooleanValueCreator(spinner);
                    valueCreators.set(
                        valueCreators.indexOf(
                            reusableValueCreators.get(key)),
                        valueCreator
                    );
                    reusableValueCreators.put(key,
                                              valueCreator);
                }
                valueCreators.add(reusableValueCreators.get(key));
                return spinner;
            }
        });
    }

    /**
     * Reads a value from a swing model object.
     * Example: If the associated model is a <code>Document</code> it reads
     * the containing text.
     */
    private static interface ValueCreator {
        public Object getValue();
    }

    /**
     * A {@link ValueCreator} for a <code>JTextComponent</code>.
     */
    private static class StringValueCreator implements ValueCreator {
        private JTextComponent text;

        public StringValueCreator(JTextComponent text) {
            this.text = text;
        }

        public Object getValue() {
            return text.getText();
        }
    }

    /**
     * A {@link ValueCreator} for a <code>JTextComponent</code>.
     */
    private static class ChoiceValueCreator implements ValueCreator {
        private JComboBox text;

        public ChoiceValueCreator(JComboBox text) {
            this.text = text;
        }

        public Object getValue() {
            return text.getSelectedItem();
        }
    }

    /**
     * A {@link ValueCreator} for a <code>JSpinner</code>.
     */
    private static class IntValueCreator implements ValueCreator {
        private JSpinner text;

        public IntValueCreator(JSpinner text) {
            this.text = text;
        }

        public Object getValue() {
            return text.getValue();
        }
    }

    /**
     * A {@link ValueCreator} for a <code>JSpinner</code>.
     */
    private static class DateValueCreator implements ValueCreator {
        private JSpinner text;

        public DateValueCreator(JSpinner text) {
            this.text = text;
        }

        public Object getValue() {
            java.util.Date date =
                ((SpinnerDateModel) text.getModel()).getDate();
            return date;
        }
    }

    /**
     * A {@link ValueCreator} for a <code>JCheckBox</code>.
     */
    private static class BooleanValueCreator implements ValueCreator {
        private JCheckBox text;

        public BooleanValueCreator(JCheckBox text) {
            this.text = text;
        }

        public Object getValue() {
            return text.isSelected();
        }
    }

    /**
     * A map to associate a single {@link DataColumn} with a
     * {@link ValueCreator} that reads the new value for this option
     * from a <code>JComponent</code>.
     * Example: The {@link DataColumn} has the type <code>Boolean.class</code>.
     * Then it has an associated <code>JCheckbox</code> in the 
     * <code>JDialog</code>. This <code>Map</code> contains the
     * {@link NewRecordPanel.ValueCreator} that reads the status of the 
     * <code>JCheckBox</code>
     * (checked or not).
     */
    private final java.util.List<ValueCreator> valueCreators =
        new ArrayList<ValueCreator>();

    private final Map<String,ValueCreator> reusableValueCreators =
        new HashMap<String,ValueCreator>();
}
