package org.kaiwitte.workbench.ui.view;

import org.kaiwitte.workbench.data.*;
import org.kaiwitte.workbench.ui.model.MetaSchemeTableModel;
import org.kaiwitte.workbench.ui.controller.SearchAction;
import org.kaiwitte.workbench.ui.controller.ReturnKeyRemapper;
import org.kaiwitte.workbench.ui.UIConfig;
import org.kaiwitte.workbench.tools.Tools;
import org.kaiwitte.tools.SelectAllListener;

import javax.swing.*;
import javax.swing.event.DocumentListener;
import javax.swing.event.DocumentEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import javax.swing.text.JTextComponent;
import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;
import java.util.HashMap;
import java.util.Map;
import java.util.ArrayList;

import static java.awt.GridBagConstraints.*;

/**
 * A <code>JPanel</code> that allows the user to search for a new record.
 */ 
public final class SearchPanel extends JPanel implements ConditionCreator {
    private final MetaSchemeTableModel tableModel;
    private final UIConfig config;
    private int row = 0;
    private final ReturnKeyRemapper remapper = 
        new ReturnKeyRemapper();
    private ConditionCreator searchByID;
    
    private final EditableFocusTraversalPolicy policy =
        new EditableFocusTraversalPolicy();

    /**
     * Creates a new instance.
     * @param model the <code>MetaSchemeTableModel</code> that defines
     * the number and type of the search fields.
     * @param config the <code>UIConfig</code>
     */ 
    public SearchPanel(MetaSchemeTableModel model, UIConfig config) {
        this.config = config;
        tableModel = model;
        setLayout(new GridBagLayout());
        addSearchFields(model);
        
        // buttons
        GridBagConstraints c = new GridBagConstraints();
        JButton search = new JButton(
            new SearchAction(config, this));
        c.gridwidth = GridBagConstraints.RELATIVE;
        c.fill = GridBagConstraints.NONE;
        c.anchor = WEST;
        c.weightx = 0.0;
        c.gridy = row;
        search.setMnemonic('s');
        add(search, c);
        remapper.remap(search);
    }

    /**
     * Returns the <code>FocusTraversalPolicy</code> that is optimised
     * for navigating in this panel
     * @return the <code>FocusTraversalPolicy</code> 
     */
    public FocusTraversalPolicy getPolicy() {
        return policy;
    }
    
    private void addSearchFields(MetaSchemeTableModel model) {
        GridBagConstraints c = new GridBagConstraints();
        
        // add search for id field
        final JCheckBox checkBox2 = new JCheckBox();
        JLabel label2 = new JLabel("ID");
        final JSpinner component2 = new JSpinner();

        c.fill = GridBagConstraints.HORIZONTAL;
        c.weightx = 0.5;
        c.gridx = 0;
        c.gridy = row;
        add(checkBox2, c);

        c.gridx = 1;
        c.gridy = row;
        add(label2, c);

        c.gridx = 2;
        c.gridy = row;
        add(component2, c);
        policy.add(component2);
        remapper.remap(component2);
        row++;
        
        searchByID = 
            new ConditionCreator() {
                public Condition createCondition() {
                    if (checkBox2.isSelected()) {
                        return ConditionFactory.rawSQLCondition(
                            tableModel.getScheme().getDataSchemes().get(0).getName()
                            + ".ID=" + component2.getValue());
                    }
                    else {
                        return ConditionFactory.rawSQLCondition(
                            tableModel.getScheme().getDataSchemes().get(0).getName()
                            + ".ID regexp '.*'");
                    }

                }
            };
        
        component2.getModel().addChangeListener(
            new ArmListener(checkBox2.getModel())
        );

        for (DataColumn column : model.getScheme().getColumns()) {
            JCheckBox checkBox = new JCheckBox();
            JLabel label = new JLabel(column.getName());
            ComponentCreator componentCreator = 
                CLASS_TO_COMPONENT.get(column.getType());
            JComponent component = componentCreator.createComponent(
                column,checkBox.getModel());

            c.fill = GridBagConstraints.HORIZONTAL;
            c.weightx = 0.5;
            c.gridx = 0;
            c.gridy = row;
            add(checkBox, c);

            c.gridx = 1;
            c.gridy = row;
            add(label, c);

            c.gridx = 2;
            c.gridy = row;
            add(component, c);
            policy.add(component);
            remapper.remap(component);

            row++;
        }
    }
    
    /**
     * Creates a {@link Condition} for the entered search criterions.
     * @return the resulting <code>Condition</code>
     */ 
    public Condition createCondition() {
        Condition ret = null;
        for (ConditionCreator c : conditionCreators.values()) {
            if (ret == null) {
                ret = c.createCondition();
            }
            else {
                ret = ConditionFactory.and(ret, c.createCondition());
            }
        }
        ret = ConditionFactory.and(ret, searchByID.createCondition());
        
        return ret;
    }
    
    public MetaSchemeTableModel getTableModel() {
        return tableModel;
    }

    /**
     * Creates a JComponent for an ???.
     * Example: Create a <code>JCheckBox</code> if the associated
     * <code>Option</code>'s type is <code>Boolean.class</code>.
     */ 
    private static interface ComponentCreator {
        public JComponent createComponent(DataColumn column, ButtonModel model);
    }
    
    /**
     * A class to enable a <code>JCheckBox</code> when the value
     * of a <code>JTextComponent</code> or a <code>JSpinner</code>
     * changes.
     */ 
    private static final class ArmListener 
            implements DocumentListener, ChangeListener, ActionListener,
                       ItemListener {
        private final ButtonModel model;

        public ArmListener(ButtonModel model) {
            this.model = model;
        }

        public void insertUpdate(DocumentEvent e) {
            model.setSelected(true);
        }

        public void removeUpdate(DocumentEvent e) {
            
        }

        public void changedUpdate(DocumentEvent e) {
            
        }

        public void stateChanged(ChangeEvent e) {
            model.setSelected(true);
        }

        public void actionPerformed(ActionEvent e) {
            model.setSelected(true);
        }

        public void itemStateChanged(ItemEvent e) {
            model.setSelected(true);
        }
    }

    /**
         * A {@link ConditionCreator} for a <code>JTextComponent</code>.
         */ 
        private static class StringConditionCreator implements ConditionCreator {
            private JTextComponent text;
            private final DataColumn column;
            private final ButtonModel model;

            public StringConditionCreator(JTextComponent text,
                                          DataColumn column,
                                          ButtonModel model) {
                this.text = text;
                text.getDocument().addDocumentListener(
                    new ArmListener(model)
                );
                this.column = column;
                this.model = model;
            }

            public Condition createCondition() {
                if (model.isSelected()) {
                    return ConditionFactory.columnEqual(column, text.getText());
                }
                else {
                    return ConditionFactory.any(column);
                }
            }
        }
    
        /**
         * A {@link ConditionCreator} for a <code>JTextComponent</code>.
         */ 
        private static class ChoiceConditionCreator implements ConditionCreator {
            private JComboBox text;
            private final DataColumn column;
            private final ButtonModel model;

            public ChoiceConditionCreator(JComboBox text,
                                          DataColumn column,
                                          ButtonModel model) {
                this.text = text;
                text.addActionListener(
                    new ArmListener(model)
                );
                this.column = column;
                this.model = model;
            }

            public Condition createCondition() {
                if (model.isSelected()) {
                    return ConditionFactory.columnEqual(
                        column,  (String) text.getSelectedItem());
                }
                else {
                    return ConditionFactory.any(column);
                }
            }
        }
    
        /**
         * A {@link ConditionCreator} for a <code>JSpinner</code>.
         */ 
        private static class IntConditionCreator implements ConditionCreator {
            private JSpinner text;
            private final DataColumn column;
            private final ButtonModel model;

            public IntConditionCreator(JSpinner text,
                                       DataColumn column,
                                       ButtonModel model) {
                this.text = text;
                text.getModel().addChangeListener(
                    new ArmListener(model)
                );
                this.column = column;
                this.model = model;
            }

            public Condition createCondition() {
                if (model.isSelected()) {
                    return ConditionFactory.columnEqual(
                        column, 
                        Integer.parseInt(text.getValue().toString()));

                }
                else {
                    return ConditionFactory.any(column);
                }
            }
        }
    
        private static class DateConditionCreator implements ConditionCreator {
            private JSpinner text;
            private final DataColumn column;
            private final ButtonModel model;

            public DateConditionCreator(JSpinner text,
                                        DataColumn column,
                                        ButtonModel model) {
                this.text = text;
                text.getModel().addChangeListener(
                    new ArmListener(model)
                );
                this.column = column;
                this.model = model;
            }

            public Condition createCondition() {
                if (model.isSelected()) {
                    java.util.Date date = 
                        ((SpinnerDateModel) text.getModel()).getDate();
                    return ConditionFactory.columnEqual(
                        column, 
                        date);

                }
                else {
                    return ConditionFactory.any(column);
                }
            }
        }
    
    private static class BooleanConditionCreator implements ConditionCreator {
        private JCheckBox text;
        private final DataColumn column;
        private final ButtonModel model;

        public BooleanConditionCreator(JCheckBox text,
                                    DataColumn column,
                                    ButtonModel model) {
            this.text = text;
            text.getModel().addItemListener(
                new ArmListener(model)
            );
            this.column = column;
            this.model = model;
        }

        public Condition createCondition() {
            if (model.isSelected()) {
                Boolean b = text.isSelected();
                return ConditionFactory.columnEqual(column, b);
            }
            else {
                return ConditionFactory.any(column);
            }
        }
    }
    
    /**
     * A map to associate a single ??? with a
     * {@link ConditionCreator} that reads the new value for this option
     * from a <code>JComponent</code>.
     * Example: The ??? has the type <code>Boolean.class</code>.
     * Then it has an associated <code>JCheckbox</code> in the 
     * <code>JDialog</code>. This <code>Map</code> contains the
     * {@link ConditionCreator} that reads the status of the <code>JCheckBox</code>
     * (checked or not).
     */ 
    private final java.util.Map<DataColumn,ConditionCreator> 
        conditionCreators = 
            new HashMap<DataColumn,ConditionCreator>();

    /**
     * A <code>Map</code> to associate a <code>Class</code> with a
     * {@link ComponentCreator} that creates the <code>JComponent</code>
     * that is best to represent the value of an instance of this 
     * <code>Class</code>.
     * Example: A <code>Boolean.class</code> is mapped to a
     * {@link ComponentCreator} that creates a <code>JCheckBox</code>.
     */ 
    private static final Map<Class,ComponentCreator> CLASS_TO_COMPONENT =
        new HashMap<Class,ComponentCreator>();
    {
        CLASS_TO_COMPONENT.put(String.class, new ComponentCreator() {
            public JComponent createComponent(DataColumn column,
                                              ButtonModel model) {
                JTextField text = new JTextField(15);
                text.addFocusListener(new SelectAllListener());
                //text.addFocusListener(new SelectAllListener());
                conditionCreators.put(column, 
                                      new StringConditionCreator(text,
                                                                 column,
                                                                 model));
                return text;
            }
        });
        CLASS_TO_COMPONENT.put(String[].class, new ComponentCreator() {
            public JComponent createComponent(DataColumn column,
                                              ButtonModel model) {
                JComboBox text = new JComboBox(
                    Tools.sequenceToArray(column.getChoices(), String[].class));
                //text.addFocusListener(new SelectAllListener());
                conditionCreators.put(column, 
                                      new ChoiceConditionCreator(text,
                                                                 column,
                                                                 model));
                return text;
            }
        });
        CLASS_TO_COMPONENT.put(Integer.class, new ComponentCreator() {
            public JComponent createComponent(DataColumn column,
                                              ButtonModel model) {
                JSpinner spinner = new JSpinner();
                //spinner.setValue(new Integer(o.createCondition()));
                conditionCreators.put(column, new IntConditionCreator(spinner,
                                                                  column,
                                                                  model));
                return spinner;
            }
        });
        CLASS_TO_COMPONENT.put(java.util.Date.class, new ComponentCreator() {
            public JComponent createComponent(DataColumn column,
                                              ButtonModel model) {
                JSpinner spinner = new JSpinner();
                spinner.setModel(new SpinnerDateModel());
                spinner.setEditor(new JSpinner.DateEditor(spinner,
                                                          "dd.MM.yyyy"));
                
                //spinner.setValue(new Integer(o.createCondition()));
                conditionCreators.put(column, new DateConditionCreator(spinner,
                                                                  column,
                                                                  model));
                return spinner;
            }
        });
        CLASS_TO_COMPONENT.put(Boolean.class, new ComponentCreator() {
            public JComponent createComponent(DataColumn column,
                                              ButtonModel model) {
                JCheckBox button = new JCheckBox();
                //button.setSelected(Boolean.parseBoolean(o.createCondition()));
                conditionCreators.put(column, new BooleanConditionCreator(button,
                                                                      column,
                                                                      model));
                return button;
            }
        });
    }
}
